import asyncio
import json
import time
import traceback
from starlette.websockets import WebSocket
from websockets import ConnectionClosed
from gallama.data_classes.realtime_client_proto import (
    SessionConfig,
    ConversationItem,
    ConversationItemMessage,
    ContentType,
    MessageContent,
    ResponseCreate,
    ConversationItemInputAudioTranscriptionComplete,
    ConversationItemCreate,
    ConversationItemTruncate,
    ConversationItemFunctionCall,
    ConversationItemFunctionCallOutput,
)
from gallama.data_classes.realtime_server_proto import (
    MessageContentServer,
    ConversationItemMessageServer,
    ConversationItemServer,
    ContentTypeServer,
    InputAudioBufferSpeechStarted,
    InputAudioBufferSpeechStopped,
    InputAudioBufferCommitted,
    ConversationItemFunctionCallServer,
    ConversationItemFunctionCallOutputServer
)
from gallama.data_classes.internal_ws import WSInterSTTResponse, WSInterSTT
from gallama.realtime.response import Response
from gallama.realtime.session_manager import SessionManager
from gallama.realtime.websocket_handler import WebSocketMessageHandler
from gallama.realtime.websocket_session import WebSocketSession
from ..dependencies_server import get_server_logger

logger = get_server_logger()

class WebSocketManager:
    """Orchestrates the overall WebSocket operations"""
    def __init__(self, session_manager: SessionManager, message_handler: WebSocketMessageHandler):
        self.session_manager = session_manager
        self.message_handler = message_handler

    async def stop_background_tasks(self, session: WebSocketSession):
        """Cancel all background tasks associated with the session."""
        if hasattr(session, 'tasks'):
            for task in session.tasks:
                task.cancel()
            session.tasks = []  # Clear the list of tasks

    async def initialize_session(
        self,
        websocket: WebSocket,
        model: str,
        api_key: str = None,
        stt_url: str = None,
        llm_url: str = None,
        tts_url: str = None,
    ) -> WebSocketSession:
        protocols = websocket.headers.get("sec-websocket-protocol", "").split(", ")
        if "openai-beta.realtime-v1" in protocols:
            await websocket.accept(subprotocol="openai-beta.realtime-v1")
        else:
            await websocket.accept()

        session_id = str(id(websocket))

        # Create session config with the model parameter
        config = SessionConfig(
            model=model,
            modalities=["text"],  # or ["text", "audio"] if you're supporting audio
            instructions="",
            streaming_transcription=True
        )

        self.message_handler.initialize(stt_url, llm_url, tts_url)

        session = self.session_manager.create_session(session_id, config)

        # Format the session data according to OpenAI's expected structure
        session_data = {
            "event_id": await session.queues.next_event(),
            "type": "session.created",
            "session": {
                **config.model_dump(exclude_none=True),  # This ensures null values aren't included
                "id": session_id  # Add the session ID explicitly
            }
        }

        await websocket.send_json(session_data)
        return session

    async def start_background_tasks(self, session: WebSocketSession, websocket: WebSocket):
        """Initialize and start background processing tasks"""

        # start connection with individual ws
        if not self.message_handler.initialize_ws:
            await self.message_handler.initialize_connection()

        session.tasks = [
            asyncio.create_task(self.process_unprocessed_queue(session, websocket)),
            asyncio.create_task(self.process_unprocess_audio_buffer(session, websocket)),
            # asyncio.create_task(self.process_audio_to_client(session, websocket))
        ]


    async def send_message_to_client(self, websocket: WebSocket, message: str):
        await websocket.send_json(message)

    async def handle_message_item(self, item: ConversationItem) -> ConversationItemServer:
        # process item
        try:
            processed_item = None
            if isinstance(item, ConversationItemMessage):
                # Create new content list for the updated items
                new_content = []

                for msg in item.content:
                    if msg.type == ContentType.INPUT_AUDIO:
                        # Create new message content with transcribed text
                        new_msg = MessageContentServer(
                            type=msg.type,
                            text=new_msg.text,
                            audio=msg.audio
                        )
                        new_content.append(new_msg)
                    else:
                        new_content.append(MessageContentServer(**msg.model_dump()))

                # Create new item with updated content
                # also convert it to server type
                processed_item = ConversationItemMessageServer(
                    id=item.id,
                    type=item.type,
                    role=item.role,
                    content=new_content,
                    status=item.status if item.status else "completed",
                )
            elif isinstance(item, ConversationItemFunctionCall):

                processed_item = ConversationItemFunctionCallServer(
                    id=item.id,
                    status="completed",
                    type=item.type,
                    call_id=item.call_id,
                    name=item.name,
                    arguments=item.arguments
                )
            elif isinstance(item, ConversationItemFunctionCallOutput):
                logger.info("FROG FROG Function call output")
                processed_item = ConversationItemFunctionCallOutputServer(
                    status="completed",
                    call_id=item.call_id,
                    output=item.output,
                )

            else:
                logger.error(f"Unknown item type: {item.type}")

            return processed_item
        except Exception as e:
            logger.error(f"Error in handle_message_item: {str(e)}\n{traceback.format_exc()}")

    async def process_unprocess_audio_buffer(self, session: WebSocketSession, websocket: WebSocket):

        try:
            ws_stt = self.message_handler.ws_stt

            while True:

                if not await ws_stt.ensure_connection():
                    raise Exception("Could not establish STT connection")

                # Receive message from STT
                message = await ws_stt.connection.recv()

                # Parse the message
                try:
                    stt_response = WSInterSTTResponse(**json.loads(message))
                    if stt_response.type == "stt.add_transcription":
                        logger.info(f"Adding transcription: {stt_response.transcription}")
                        await session.queues.append_transcription(stt_response.transcription)

                    elif stt_response.type == "stt.vad_speech_start":
                        logger.info(f"websocket_manager: VAD speech start: {stt_response}")
                        # get new item id
                        # session.queues.vad_item_id = await session.queues.next_item()
                        session.queues.vad_item_id_next_in_queue = await session.queues.next_item()

                        # track the current time for llm, offset by prefix padding from vad
                        session.queues.speech_start_time_in_queue = time.time()
                        if session.config.turn_detection and session.config.turn_detection.prefix_padding_ms and session.config.turn_detection.prefix_padding_ms>0:
                            session.queues.speech_start_time_in_queue -= session.config.turn_detection.prefix_padding_ms / 1000

                        # send client
                        await websocket.send_json(InputAudioBufferSpeechStarted(
                            event_id=await session.queues.next_event(),
                            audio_start_ms=stt_response.vad_timestamp_ms,
                            item_id=session.queues.vad_item_id_next_in_queue
                        ).model_dump())

                        # update tracking
                        session.queues.speech_start = stt_response.vad_timestamp_ms
                        session.queues.speech_end = None

                        # ensure transcription is reset set to
                        logger.info(f"Reset transcription")
                        await session.queues.clear_transcription()

                    elif stt_response.type == "stt.vad_speech_end":
                        logger.info(f"websocket_manager: VAD speech end: {stt_response}")
                        # swap the item id in
                        session.queues.vad_item_id = session.queues.vad_item_id_next_in_queue
                        session.queues.vad_item_id_next_in_queue = None

                        session.queues.speech_end_time = time.time()
                        session.queues.speech_start_time = session.queues.speech_start_time_in_queue
                        session.queues.speech_start_time_in_queue = None

                        await websocket.send_json(InputAudioBufferSpeechStopped(
                            event_id=await session.queues.next_event(),
                            audio_end_ms=stt_response.vad_timestamp_ms,
                            item_id=session.queues.vad_item_id
                        ).model_dump())

                        # update tracking
                        session.queues.speech_end = stt_response.vad_timestamp_ms

                        # handling internal audio commit
                        await session.queues.commit_unprocessed_audio(ws_stt=ws_stt, item_id=session.queues.vad_item_id, skip_sound_done=True)

                        # send user committed msg
                        await websocket.send_json(InputAudioBufferCommitted(**{
                            "event_id": await session.queues.next_event(),
                            "type": "input_audio_buffer.committed",
                            "previous_item_id": await session.queues.current_item_id(),
                            "item_id": session.queues.vad_item_id
                        }).model_dump())

                        # trigger response
                        response_event = ResponseCreate(
                            id=await session.queues.next_resp(),
                            speech_start_time=session.queues.speech_start_time,
                            speech_end_time=session.queues.speech_end_time,
                        )
                        await session.queues.unprocessed.put(response_event)
                    elif stt_response.type == "stt.transcription_complete":
                        logger.info(f"Received Transcription complete from TTS")
                        await session.queues.mark_transcription_done()
                        # break     # not break as this is meant to run forever
                    elif stt_response.type == "stt.config_updated":
                        logger.info(f"TTS acknowledge config updated")
                    elif stt_response.type == "stt.buffer_cleared":
                        # clear transcription
                        await session.queues.clear_transcription()
                    else:
                        logger.error(f"Unknown stt message type: {stt_response.type}")
                except asyncio.TimeoutError:
                    continue
                except json.JSONDecodeError:
                    logger.error(f"Failed to parse STT message: {message}")
                    continue
                except ConnectionClosed:
                    logger.error("Stt websocket connection closed unexpectedly")
                    await ws_stt.ensure_connection()
                    continue
        except Exception as e:
            logger.error(f"Unexpected error occurred in process_unprocess_audio_buffer: {str(e)}\n{traceback.format_exc()}")
            raise e


    async def process_unprocessed_queue(self, session: WebSocketSession, ws_client: WebSocket):
        """Process items in unprocessed queue one at a time"""
        while True:
            item: ConversationItemCreate | ResponseCreate = await session.queues.unprocessed.get()
            logger.info(f"Processing item: {item}")
            try:
                if isinstance(item, ConversationItemCreate):
                    item_to_create = await self.handle_message_item(item.item)
                    await session.queues.update_conversation_item_ordered_dict(
                        ws_client=ws_client,
                        ws_llm=self.message_handler.ws_llm,
                        item=item_to_create,
                    )
                elif isinstance(item, ConversationItemTruncate):
                    logger.info(f"truncating conversation item")
                    # check if need to factor in prefix padding ms
                    offset_timing = 0
                    if session.config.turn_detection and session.config.turn_detection.prefix_padding_ms:
                        offset_timing = session.config.turn_detection.prefix_padding_ms

                    await session.queues.truncate_conversation_item(
                        ws_client=ws_client,
                        ws_llm=self.message_handler.ws_llm,
                        event=item,
                        user_interrupt_token=session.config.user_interrupt_token,
                        offset_timing=offset_timing,
                    )
                elif isinstance(item, ResponseCreate):
                    # if there is audio commited, create an item for it

                    modalities = "audio" if "audio" in session.config.modalities else "text"
                    logger.info(f"modalities: {modalities}")
                    logger.info(f"session.config.modalities: {session.config.modalities}")
                    # Process transcription and response
                    if modalities == "audio":
                        # wait until transcription is done
                        if session.config.turn_detection and not session.config.turn_detection.create_response:
                            transcription_done, vad_item_id = await session.queues.wait_for_transcription_done(use_vad=True)
                        else:
                            transcription_done, vad_item_id = await session.queues.wait_for_transcription_done(use_vad=False)

                        logger.info(f"transcription_done: {transcription_done}")
                        logger.info(f"vad_item_id: {vad_item_id}")
                        logger.info(f"session.queues.transcript_buffer: {session.queues.transcript_buffer}")

                        if vad_item_id and transcription_done and session.queues.transcript_buffer:

                            user_audio_item = ConversationItemMessageServer(
                                id=session.queues.vad_item_id,  # commit_unprocessed_audio save the id
                                type="message",
                                role="user",
                                status="completed",
                                content=[
                                    MessageContentServer(
                                        type="input_audio",
                                        audio=session.queues.audio_buffer,
                                        transcript=session.queues.transcript_buffer
                                    )
                                ]
                            )

                            await session.queues.update_conversation_item_ordered_dict(
                                ws_client=ws_client,
                                ws_llm=self.message_handler.ws_llm,
                                item=user_audio_item
                            )

                            await ws_client.send_json(ConversationItemInputAudioTranscriptionComplete(
                                event_id=await session.queues.next_event(),
                                type="conversation.item.input_audio_transcription.completed",
                                item_id=session.queues.vad_item_id,
                                content_index=0,
                                transcript=session.queues.transcript_buffer
                            ).model_dump())

                    # get response counter
                    response_id = await session.queues.next_resp()

                    # create new response object that will manage this response
                    response = Response(
                        session=session,
                        ws_client=ws_client,
                        ws_stt=self.message_handler.ws_stt,
                        ws_llm=self.message_handler.ws_llm,
                        ws_tts=self.message_handler.ws_tts,
                        response_id=response_id,
                        event_id_generator=session.queues.next_event
                    )

                    # set the current response for cancellation
                    async with session.current_response_lock:
                        logger.info(f"-------------------Set Current response")
                        session.current_response = response

                    # if there is no timing in the response create (non vad mode)
                    # using timing of audio as timing for video
                    if not item.speech_start_time and not item.speech_end_time:
                        if session.queues.audio_buffer.any():
                            item.speech_end_time = time.time()
                            item.speech_start_time = item.speech_end_time - session.queues.calculate_audio_duration()


                    # at this point stt already completed, and text send to the llm
                    await response.response_initialize()
                    await response.response_start(response_create_request=item, session_config=session.config)

                    # update conversation list with assistant answer
                    logger.info("update conversation list with assistant answer")
                    await session.queues.update_conversation_item_with_assistant_response(
                        item_id=response.item_id,
                        _type=modalities,
                        text=response.text or response.transcription,   # for audio item, fill text with transcription
                        transcript=response.transcription,
                        audio=response.audio,
                        function_calling_name=response.function_calling_name,
                        function_calling_arguments=response.function_calling_arguments,
                        function_calling_tool_id=response.function_calling_tool_id,
                        ws_llm=self.message_handler.ws_llm,
                    )

                    await session.queues.reset_after_response()

                    logger.info(f"---------------------Response created: {response_id}")

                    # remove current response
                    async with session.current_response_lock:
                        logger.info(f"-------------------Reset current response")
                        session.current_response = None

                    # clear vad_item_id
                    session.queues.vad_item_id = None


            except Exception as e:
                logger.error(f"Error processing queue item: {str(e)}")
            finally:
                session.queues.unprocessed.task_done()



from fastapi import Request, APIRouter
from gallama.data_classes import (
    ModelObjectResponse,
    ModelObject,
    ModelSpec
)
from ..logger import logger
from ..dependencies import get_model_manager

router = APIRouter(prefix="", tags=["model_management"])


@router.post("/load_model")
def load_model(model_spec: ModelSpec):
    """
    model_spec is model specification coming from cli
    it might not have all the properties required for the model to be loaded
    the config_manager below contain all the models properties
    """
    model_manager = get_model_manager()
    model_manager.load_model(model_spec)


@router.get("/v1/models")
async def get_models(request: Request):
    model_manager = get_model_manager()

    data = []
    for model_name in model_manager.llm_dict.keys():
        data.append(ModelObject(id=model_name))
    for model_name in model_manager.stt_dict.keys():
        data.append(ModelObject(id=model_name))
    for model_name in model_manager.tts_dict.keys():
        data.append(ModelObject(id=model_name))
    for model_name in model_manager.embedding_dict.keys():
        data.append(ModelObject(id=model_name))
    return ModelObjectResponse(data=data)

# @router.post("/delete_model")
# def delete_model(model_name: str):
#
#     if model_name in llm_dict.keys():
#         del llm_dict[model_name]
#
#         # Clear CUDA cache if using GPU
#         if torch.cuda.is_available():
#             torch.cuda.empty_cache()
#
#         # Run garbage collection
#         gc.collect()
#
#         logger.info("Deleted model: " + model_name)


@router.get("/status")
async def get_status():
    model_manager = get_model_manager()
    logger.info("Status endpoint called")
    return {"status": "ready" if model_manager.model_ready else "loading"}

# Add a health check endpoint
@router.get("/health")
async def health_check():
    return {"status": "healthy"}


import os

inp_text = os.environ.get("inp_text")
exp_name = os.environ.get("exp_name")
i_part = os.environ.get("i_part")
all_parts = os.environ.get("all_parts")
if "_CUDA_VISIBLE_DEVICES" in os.environ:
     os.environ["CUDA_VISIBLE_DEVICES"] = os.environ["_CUDA_VISIBLE_DEVICES"]
opt_dir = os.environ.get("opt_dir")
pretrained_s2G = os.environ.get("pretrained_s2G")
s2config_path = os.environ.get("s2config_path")
version=os.environ.get("version","v2")
import torch
is_half = eval(os.environ.get("is_half", "True")) and torch.cuda.is_available()
import math, traceback
import multiprocessing
import sys, pdb

now_dir = os.getcwd()
sys.path.append(now_dir)
from random import shuffle
import torch.multiprocessing as mp
from glob import glob
from tqdm import tqdm
import logging, librosa, utils
from module.models import SynthesizerTrn
from tools.my_utils import clean_path
logging.getLogger("numba").setLevel(logging.WARNING)
# from config import pretrained_s2G

# inp_text=sys.argv[1]
# exp_name=sys.argv[2]
# i_part=sys.argv[3]
# all_parts=sys.argv[4]
# os.environ["CUDA_VISIBLE_DEVICES"]=sys.argv[5]
# opt_dir="/data/docker/liujing04/gpt-vits/fine_tune_dataset/%s"%exp_name

if os.path.exists(pretrained_s2G):...
else:raise FileNotFoundError(pretrained_s2G)

hubert_dir = "%s/4-cnhubert" % (opt_dir)
semantic_path = "%s/6-name2semantic-%s.tsv" % (opt_dir, i_part)
if os.path.exists(semantic_path) == False:
    os.makedirs(opt_dir, exist_ok=True)

    if torch.cuda.is_available():
        device = "cuda"
    # elif torch.backends.mps.is_available():
    #     device = "mps"
    else:
        device = "cpu"
    hps = utils.get_hparams_from_file(s2config_path)
    vq_model = SynthesizerTrn(
        hps.data.filter_length // 2 + 1,
        hps.train.segment_size // hps.data.hop_length,
        n_speakers=hps.data.n_speakers,
        version=version,
        **hps.model
    )
    if is_half == True:
        vq_model = vq_model.half().to(device)
    else:
        vq_model = vq_model.to(device)
    vq_model.eval()
    # utils.load_checkpoint(utils.latest_checkpoint_path(hps.s2_ckpt_dir, "G_*.pth"), vq_model, None, True)
    # utils.load_checkpoint(pretrained_s2G, vq_model, None, True)
    print(
        vq_model.load_state_dict(
            torch.load(pretrained_s2G, map_location="cpu")["weight"], strict=False
        )
    )

    def name2go(wav_name, lines):
        hubert_path = "%s/%s.pt" % (hubert_dir, wav_name)
        if os.path.exists(hubert_path) == False:
            return
        ssl_content = torch.load(hubert_path, map_location="cpu")
        if is_half == True:
            ssl_content = ssl_content.half().to(device)
        else:
            ssl_content = ssl_content.to(device)
        codes = vq_model.extract_latent(ssl_content)
        semantic = " ".join([str(i) for i in codes[0, 0, :].tolist()])
        lines.append("%s\t%s" % (wav_name, semantic))

    with open(inp_text, "r", encoding="utf8") as f:
        lines = f.read().strip("\n").split("\n")

    lines1 = []
    for line in lines[int(i_part) :: int(all_parts)]:
        # print(line)
        try:
            # wav_name,text=line.split("\t")
            wav_name, spk_name, language, text = line.split("|")
            wav_name=clean_path(wav_name)
            wav_name = os.path.basename(wav_name)
            # name2go(name,lines1)
            name2go(wav_name, lines1)
        except:
            print(line, traceback.format_exc())
    with open(semantic_path, "w", encoding="utf8") as f:
        f.write("\n".join(lines1))



import argparse,os
import copy
import json
import os
import uuid

try:
    import gradio.analytics as analytics
    analytics.version_check = lambda:None
except:...

import librosa
import gradio as gr
import numpy as np
import soundfile

g_json_key_text = ""
g_json_key_path = ""
g_load_file = ""
g_load_format = ""

g_max_json_index = 0
g_index = 0
g_batch = 10
g_text_list = []
g_audio_list = []
g_checkbox_list = []
g_data_json = []


def reload_data(index, batch):
    global g_index
    g_index = index
    global g_batch
    g_batch = batch
    datas = g_data_json[index:index+batch]
    output = []
    for d in datas:
        output.append(
            {
                g_json_key_text: d[g_json_key_text],
                g_json_key_path: d[g_json_key_path]
            }
        )
    return output


def b_change_index(index, batch):
    global g_index, g_batch
    g_index, g_batch = index, batch
    datas = reload_data(index, batch)
    output = []
    for i , _ in enumerate(datas):
        output.append(
            # gr.Textbox(
            #     label=f"Text {i+index}",
            #     value=_[g_json_key_text]#text
            # )
            {
                "__type__":"update",
                "label":f"Text {i+index}",
                "value":_[g_json_key_text]
            }
        )
    for _ in range(g_batch - len(datas)):
        output.append(
            # gr.Textbox(
            #     label=f"Text",
            #     value=""
            # )
            {
                "__type__": "update",
                "label": f"Text",
                "value": ""
            }
        )
    for _ in datas:
        output.append(_[g_json_key_path])
    for _ in range(g_batch - len(datas)):
        output.append(None)
    for _ in range(g_batch):
        output.append(False)
    return output


def b_next_index(index, batch):
    b_save_file()
    if (index + batch) <= g_max_json_index:
        return index + batch , *b_change_index(index + batch, batch)
    else:
        return index, *b_change_index(index, batch)


def b_previous_index(index, batch):
    b_save_file()
    if (index - batch) >= 0:
        return index - batch , *b_change_index(index - batch, batch)
    else:
        return 0, *b_change_index(0, batch)


def b_submit_change(*text_list):
    global g_data_json
    change = False
    for i, new_text in enumerate(text_list):
        if g_index + i <= g_max_json_index:
            new_text = new_text.strip()+' '
            if (g_data_json[g_index + i][g_json_key_text] != new_text):
                g_data_json[g_index + i][g_json_key_text] = new_text
                change = True
    if change:
        b_save_file()
    return g_index, *b_change_index(g_index, g_batch)


def b_delete_audio(*checkbox_list):
    global g_data_json, g_index, g_max_json_index
    b_save_file()
    change = False
    for i, checkbox in reversed(list(enumerate(checkbox_list))):
        if g_index + i < len(g_data_json):
            if (checkbox == True):
                g_data_json.pop(g_index + i)
                change = True
    
    g_max_json_index = len(g_data_json)-1
    if g_index > g_max_json_index:
        g_index = g_max_json_index
        g_index = g_index if g_index >= 0 else 0
    if change:
        b_save_file()
    # return gr.Slider(value=g_index, maximum=(g_max_json_index if g_max_json_index>=0 else 0)), *b_change_index(g_index, g_batch)
    return {"value":g_index,"__type__":"update","maximum":(g_max_json_index if g_max_json_index>=0 else 0)},*b_change_index(g_index, g_batch)


def b_invert_selection(*checkbox_list):
    new_list = [not item if item is True else True for item in checkbox_list]
    return new_list


def get_next_path(filename):
    base_dir = os.path.dirname(filename)
    base_name = os.path.splitext(os.path.basename(filename))[0]
    for i in range(100):
        new_path = os.path.join(base_dir, f"{base_name}_{str(i).zfill(2)}.wav")
        if not os.path.exists(new_path) :
            return new_path
    return os.path.join(base_dir, f'{str(uuid.uuid4())}.wav')


def b_audio_split(audio_breakpoint, *checkbox_list):
    global g_data_json , g_max_json_index
    checked_index = []
    for i, checkbox in enumerate(checkbox_list):
        if (checkbox == True and g_index+i < len(g_data_json)):
            checked_index.append(g_index + i)
    if len(checked_index) == 1 :
        index = checked_index[0]
        audio_json = copy.deepcopy(g_data_json[index])
        path = audio_json[g_json_key_path]
        data, sample_rate = librosa.load(path, sr=None, mono=True)
        audio_maxframe = len(data)
        break_frame = int(audio_breakpoint * sample_rate)

        if (break_frame >= 1 and break_frame < audio_maxframe):
            audio_first = data[0:break_frame]
            audio_second = data[break_frame:]
            nextpath = get_next_path(path)
            soundfile.write(nextpath, audio_second, sample_rate)
            soundfile.write(path, audio_first, sample_rate)
            g_data_json.insert(index + 1, audio_json)
            g_data_json[index + 1][g_json_key_path] = nextpath
            b_save_file()

    g_max_json_index = len(g_data_json) - 1
    # return gr.Slider(value=g_index, maximum=g_max_json_index), *b_change_index(g_index, g_batch)
    return {"value":g_index,"maximum":g_max_json_index,"__type__":"update"}, *b_change_index(g_index, g_batch)

def b_merge_audio(interval_r, *checkbox_list):
    global g_data_json , g_max_json_index
    b_save_file()
    checked_index = []
    audios_path = []
    audios_text = []
    for i, checkbox in enumerate(checkbox_list):
        if (checkbox == True and g_index+i < len(g_data_json)):
            checked_index.append(g_index + i)
            
    if (len(checked_index)>1):
        for i in checked_index:
            audios_path.append(g_data_json[i][g_json_key_path])
            audios_text.append(g_data_json[i][g_json_key_text])
        for i in reversed(checked_index[1:]):
            g_data_json.pop(i)

        base_index = checked_index[0]
        base_path = audios_path[0]
        g_data_json[base_index][g_json_key_text] = "".join(audios_text)

        audio_list = []
        l_sample_rate = None
        for i, path in enumerate(audios_path):
            data, sample_rate = librosa.load(path, sr=l_sample_rate, mono=True)
            l_sample_rate = sample_rate
            if (i > 0):
                silence = np.zeros(int(l_sample_rate * interval_r))
                audio_list.append(silence)

            audio_list.append(data)

        audio_concat = np.concatenate(audio_list)

        soundfile.write(base_path, audio_concat, l_sample_rate)

        b_save_file()
    
    g_max_json_index = len(g_data_json) - 1
    
    # return gr.Slider(value=g_index, maximum=g_max_json_index), *b_change_index(g_index, g_batch)
    return {"value":g_index,"maximum":g_max_json_index,"__type__":"update"}, *b_change_index(g_index, g_batch)


def b_save_json():
    with open(g_load_file,'w', encoding="utf-8") as file:
        for data in g_data_json:
            file.write(f'{json.dumps(data, ensure_ascii = False)}\n')


def b_save_list():
    with open(g_load_file,'w', encoding="utf-8") as file:
        for data in g_data_json:
            wav_path = data["wav_path"]
            speaker_name = data["speaker_name"]
            language = data["language"]
            text = data["text"]
            file.write(f"{wav_path}|{speaker_name}|{language}|{text}".strip()+'\n')


def b_load_json():
    global g_data_json, g_max_json_index
    with open(g_load_file, 'r', encoding="utf-8") as file:
        g_data_json = file.readlines()
        g_data_json = [json.loads(line) for line in g_data_json]
        g_max_json_index = len(g_data_json) - 1


def b_load_list():
    global g_data_json, g_max_json_index
    with open(g_load_file, 'r', encoding="utf-8") as source:
        data_list = source.readlines()
        for _ in data_list:
            data = _.split('|')
            if (len(data) == 4):
                wav_path, speaker_name, language, text = data
                g_data_json.append(
                        {
                            'wav_path':wav_path,
                            'speaker_name':speaker_name,
                            'language':language,
                            'text':text.strip()
                        }
                )
            else:
                print("error line:", data)
        g_max_json_index = len(g_data_json) - 1


def b_save_file():
    if g_load_format == "json":
        b_save_json()
    elif g_load_format == "list":
        b_save_list()


def b_load_file():
    if g_load_format == "json":
        b_load_json()
    elif g_load_format == "list":
        b_load_list()


def set_global(load_json, load_list, json_key_text, json_key_path, batch):
    global g_json_key_text, g_json_key_path, g_load_file, g_load_format, g_batch

    g_batch = int(batch)
    
    if (load_json != "None"):
        g_load_format = "json"
        g_load_file = load_json
    elif (load_list != "None"):
        g_load_format = "list"
        g_load_file = load_list
    else:
        g_load_format = "list"
        g_load_file = "demo.list"
        
    g_json_key_text = json_key_text
    g_json_key_path = json_key_path

    b_load_file()


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Process some integers.')
    parser.add_argument('--load_json', default="None", help='source file, like demo.json')
    parser.add_argument('--is_share', default="False", help='whether webui is_share=True')
    parser.add_argument('--load_list', default="None", help='source file, like demo.list')
    parser.add_argument('--webui_port_subfix', default=9871, help='source file, like demo.list')
    parser.add_argument('--json_key_text', default="text", help='the text key name in json, Default: text')
    parser.add_argument('--json_key_path', default="wav_path", help='the path key name in json, Default: wav_path')
    parser.add_argument('--g_batch', default=10, help='max number g_batch wav to display, Default: 10')

    args = parser.parse_args()

    set_global(args.load_json, args.load_list, args.json_key_text, args.json_key_path, args.g_batch)
    
    with gr.Blocks() as demo:

        with gr.Row():
            btn_change_index = gr.Button("Change Index")
            btn_submit_change = gr.Button("Submit Text")
            btn_merge_audio = gr.Button("Merge Audio")
            btn_delete_audio = gr.Button("Delete Audio")
            btn_previous_index = gr.Button("Previous Index")
            btn_next_index = gr.Button("Next Index")
            
        with gr.Row():
            index_slider = gr.Slider(
                    minimum=0, maximum=g_max_json_index, value=g_index, step=1, label="Index", scale=3
            )
            splitpoint_slider = gr.Slider(
                    minimum=0, maximum=120.0, value=0, step=0.1, label="Audio Split Point(s)", scale=3
            )
            btn_audio_split = gr.Button("Split Audio", scale=1)
            btn_save_json = gr.Button("Save File", visible=True, scale=1)
            btn_invert_selection = gr.Button("Invert Selection", scale=1)
        
        with gr.Row():
            with gr.Column():
                for _ in range(0,g_batch):
                    with gr.Row():
                        text = gr.Textbox(
                            label = "Text",
                            visible = True,
                            scale=5
                        )
                        audio_output = gr.Audio(
                            label="Output Audio",
                            visible = True,
                            scale=5
                        )
                        audio_check = gr.Checkbox(
                            label="Yes",
                            show_label = True,
                            info = "Choose Audio",
                            scale=1
                        )
                        g_text_list.append(text)
                        g_audio_list.append(audio_output)
                        g_checkbox_list.append(audio_check)



        with gr.Row():
            batchsize_slider = gr.Slider(
                    minimum=1, maximum=g_batch, value=g_batch, step=1, label="Batch Size", scale=3, interactive=False
            )
            interval_slider = gr.Slider(
                    minimum=0, maximum=2, value=0, step=0.01, label="Interval", scale=3
            )
            btn_theme_dark = gr.Button("Light Theme", link="?__theme=light", scale=1)
            btn_theme_light = gr.Button("Dark Theme", link="?__theme=dark", scale=1)
        
        btn_change_index.click(
            b_change_index,
            inputs=[
                index_slider,
                batchsize_slider,
            ],
            outputs=[
                *g_text_list,
                *g_audio_list,
                *g_checkbox_list
            ],
        )

        
        btn_submit_change.click(
            b_submit_change,
            inputs=[
                *g_text_list,
            ],
            outputs=[
                index_slider,
                *g_text_list,
                *g_audio_list,
                *g_checkbox_list
            ],
        )

        btn_previous_index.click(
            b_previous_index,
            inputs=[
                index_slider,
                batchsize_slider,
            ],
            outputs=[
                index_slider,
                *g_text_list,
                *g_audio_list,
                *g_checkbox_list
            ],
        )
        
        btn_next_index.click(
            b_next_index,
            inputs=[
                index_slider,
                batchsize_slider,
            ],
            outputs=[
                index_slider,
                *g_text_list,
                *g_audio_list,
                *g_checkbox_list
            ],
        )

        btn_delete_audio.click(
            b_delete_audio,
            inputs=[
                *g_checkbox_list
            ],
            outputs=[
                index_slider,
                *g_text_list,
                *g_audio_list,
                *g_checkbox_list
            ]
        )

        btn_merge_audio.click(
            b_merge_audio,
            inputs=[
                interval_slider,
                *g_checkbox_list
            ],
            outputs=[
                index_slider,
                *g_text_list,
                *g_audio_list,
                *g_checkbox_list
            ]
        )

        btn_audio_split.click(
            b_audio_split,
            inputs=[
                splitpoint_slider,
                *g_checkbox_list
            ],
            outputs=[
                index_slider,
                *g_text_list,
                *g_audio_list,
                *g_checkbox_list
            ]
        )

        btn_invert_selection.click(
            b_invert_selection,
            inputs=[
                *g_checkbox_list
            ],
            outputs=[
                *g_checkbox_list
            ]
        )

        btn_save_json.click(
            b_save_file
        )

        demo.load(
            b_change_index,
            inputs=[
                index_slider,
                batchsize_slider,
            ],
            outputs=[
                *g_text_list,
                *g_audio_list,
                *g_checkbox_list
            ],
        )
        
    demo.launch(
        server_name="0.0.0.0",
        inbrowser=True,
        quiet=True,
        share=eval(args.is_share),
        server_port=int(args.webui_port_subfix)
    )


# Copyright (c) 2021 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Rules to verbalize numbers into Chinese characters.
https://zh.wikipedia.org/wiki/中文数字#現代中文
"""
import re
from collections import OrderedDict
from typing import List

DIGITS = {str(i): tran for i, tran in enumerate('零一二三四五六七八九')}
UNITS = OrderedDict({
    1: '十',
    2: '百',
    3: '千',
    4: '万',
    8: '亿',
})

COM_QUANTIFIERS = '(处|台|架|枚|趟|幅|平|方|堵|间|床|株|批|项|例|列|篇|栋|注|亩|封|艘|把|目|套|段|人|所|朵|匹|张|座|回|场|尾|条|个|首|阙|阵|网|炮|顶|丘|棵|只|支|袭|辆|挑|担|颗|壳|窠|曲|墙|群|腔|砣|座|客|贯|扎|捆|刀|令|打|手|罗|坡|山|岭|江|溪|钟|队|单|双|对|出|口|头|脚|板|跳|枝|件|贴|针|线|管|名|位|身|堂|课|本|页|家|户|层|丝|毫|厘|分|钱|两|斤|担|铢|石|钧|锱|忽|(千|毫|微)克|毫|厘|(公)分|分|寸|尺|丈|里|寻|常|铺|程|(千|分|厘|毫|微)米|米|撮|勺|合|升|斗|石|盘|碗|碟|叠|桶|笼|盆|盒|杯|钟|斛|锅|簋|篮|盘|桶|罐|瓶|壶|卮|盏|箩|箱|煲|啖|袋|钵|年|月|日|季|刻|时|周|天|秒|分|小时|旬|纪|岁|世|更|夜|春|夏|秋|冬|代|伏|辈|丸|泡|粒|颗|幢|堆|条|根|支|道|面|片|张|颗|块|元|(亿|千万|百万|万|千|百)|(亿|千万|百万|万|千|百|美|)元|(亿|千万|百万|万|千|百|十|)吨|(亿|千万|百万|万|千|百|)块|角|毛|分)'

# 分数表达式
RE_FRAC = re.compile(r'(-?)(\d+)/(\d+)')


def replace_frac(match) -> str:
    """
    Args:
        match (re.Match)
    Returns:
        str
    """
    sign = match.group(1)
    nominator = match.group(2)
    denominator = match.group(3)
    sign: str = "负" if sign else ""
    nominator: str = num2str(nominator)
    denominator: str = num2str(denominator)
    result = f"{sign}{denominator}分之{nominator}"
    return result


# 百分数表达式
RE_PERCENTAGE = re.compile(r'(-?)(\d+(\.\d+)?)%')


def replace_percentage(match) -> str:
    """
    Args:
        match (re.Match)
    Returns:
        str
    """
    sign = match.group(1)
    percent = match.group(2)
    sign: str = "负" if sign else ""
    percent: str = num2str(percent)
    result = f"{sign}百分之{percent}"
    return result


# 整数表达式
# 带负号的整数 -10
RE_INTEGER = re.compile(r'(-)' r'(\d+)')


def replace_negative_num(match) -> str:
    """
    Args:
        match (re.Match)
    Returns:
        str
    """
    sign = match.group(1)
    number = match.group(2)
    sign: str = "负" if sign else ""
    number: str = num2str(number)
    result = f"{sign}{number}"
    return result


# 编号-无符号整形
# 00078
RE_DEFAULT_NUM = re.compile(r'\d{3}\d*')


def replace_default_num(match):
    """
    Args:
        match (re.Match)
    Returns:
        str
    """
    number = match.group(0)
    return verbalize_digit(number, alt_one=True)


# 加减乘除
# RE_ASMD = re.compile(
#     r'((-?)((\d+)(\.\d+)?)|(\.(\d+)))([\+\-\×÷=])((-?)((\d+)(\.\d+)?)|(\.(\d+)))')
RE_ASMD = re.compile(
    r'((-?)((\d+)(\.\d+)?[⁰¹²³⁴⁵⁶⁷⁸⁹ˣʸⁿ]*)|(\.\d+[⁰¹²³⁴⁵⁶⁷⁸⁹ˣʸⁿ]*)|([A-Za-z][⁰¹²³⁴⁵⁶⁷⁸⁹ˣʸⁿ]*))([\+\-\×÷=])((-?)((\d+)(\.\d+)?[⁰¹²³⁴⁵⁶⁷⁸⁹ˣʸⁿ]*)|(\.\d+[⁰¹²³⁴⁵⁶⁷⁸⁹ˣʸⁿ]*)|([A-Za-z][⁰¹²³⁴⁵⁶⁷⁸⁹ˣʸⁿ]*))')

asmd_map = {
    '+': '加',
    '-': '减',
    '×': '乘',
    '÷': '除',
    '=': '等于'
}

def replace_asmd(match) -> str:
    """
    Args:
        match (re.Match)
    Returns:
        str
    """
    result = match.group(1) + asmd_map[match.group(8)] + match.group(9)
    return result


# 次方专项
RE_POWER = re.compile(r'[⁰¹²³⁴⁵⁶⁷⁸⁹ˣʸⁿ]+')

power_map = {
    '⁰': '0',
    '¹': '1',
    '²': '2',
    '³': '3',
    '⁴': '4',
    '⁵': '5',
    '⁶': '6',
    '⁷': '7',
    '⁸': '8',
    '⁹': '9',
    'ˣ': 'x',
    'ʸ': 'y',
    'ⁿ': 'n'
}

def replace_power(match) -> str:
    """
    Args:
        match (re.Match)
    Returns:
        str
    """
    power_num = ""
    for m in match.group(0):
        power_num += power_map[m]
    result = "的" + power_num + "次方"
    return result


# 数字表达式
# 纯小数
RE_DECIMAL_NUM = re.compile(r'(-?)((\d+)(\.\d+))' r'|(\.(\d+))')
# 正整数 + 量词
RE_POSITIVE_QUANTIFIERS = re.compile(r"(\d+)([多余几\+])?" + COM_QUANTIFIERS)
RE_NUMBER = re.compile(r'(-?)((\d+)(\.\d+)?)' r'|(\.(\d+))')


def replace_positive_quantifier(match) -> str:
    """
    Args:
        match (re.Match)
    Returns:
        str
    """
    number = match.group(1)
    match_2 = match.group(2)
    if match_2 == "+":
        match_2 = "多"
    match_2: str = match_2 if match_2 else ""
    quantifiers: str = match.group(3)
    number: str = num2str(number)
    number = "两" if number == "二" else number
    result = f"{number}{match_2}{quantifiers}"
    return result


def replace_number(match) -> str:
    """
    Args:
        match (re.Match)
    Returns:
        str
    """
    sign = match.group(1)
    number = match.group(2)
    pure_decimal = match.group(5)
    if pure_decimal:
        result = num2str(pure_decimal)
    else:
        sign: str = "负" if sign else ""
        number: str = num2str(number)
        result = f"{sign}{number}"
    return result


# 范围表达式
# match.group(1) and match.group(8) are copy from RE_NUMBER

RE_RANGE = re.compile(
    r"""
    (?<![\d\+\-\×÷=])      # 使用反向前瞻以确保数字范围之前没有其他数字和操作符
    ((-?)((\d+)(\.\d+)?))  # 匹配范围起始的负数或正数（整数或小数）
    [-~]                   # 匹配范围分隔符
    ((-?)((\d+)(\.\d+)?))  # 匹配范围结束的负数或正数（整数或小数）
    (?![\d\+\-\×÷=])       # 使用正向前瞻以确保数字范围之后没有其他数字和操作符
    """, re.VERBOSE)


def replace_range(match) -> str:
    """
    Args:
        match (re.Match)
    Returns:
        str
    """
    first, second = match.group(1), match.group(6)
    first = RE_NUMBER.sub(replace_number, first)
    second = RE_NUMBER.sub(replace_number, second)
    result = f"{first}到{second}"
    return result


# ~至表达式
RE_TO_RANGE = re.compile(
    r'((-?)((\d+)(\.\d+)?)|(\.(\d+)))(%|°C|℃|度|摄氏度|cm2|cm²|cm3|cm³|cm|db|ds|kg|km|m2|m²|m³|m3|ml|m|mm|s)[~]((-?)((\d+)(\.\d+)?)|(\.(\d+)))(%|°C|℃|度|摄氏度|cm2|cm²|cm3|cm³|cm|db|ds|kg|km|m2|m²|m³|m3|ml|m|mm|s)')

def replace_to_range(match) -> str:
    """
    Args:
        match (re.Match)
    Returns:
        str
    """
    result = match.group(0).replace('~', '至')
    return result


def _get_value(value_string: str, use_zero: bool=True) -> List[str]:
    stripped = value_string.lstrip('0')
    if len(stripped) == 0:
        return []
    elif len(stripped) == 1:
        if use_zero and len(stripped) < len(value_string):
            return [DIGITS['0'], DIGITS[stripped]]
        else:
            return [DIGITS[stripped]]
    else:
        largest_unit = next(
            power for power in reversed(UNITS.keys()) if power < len(stripped))
        first_part = value_string[:-largest_unit]
        second_part = value_string[-largest_unit:]
        return _get_value(first_part) + [UNITS[largest_unit]] + _get_value(
            second_part)


def verbalize_cardinal(value_string: str) -> str:
    if not value_string:
        return ''

    # 000 -> '零' , 0 -> '零'
    value_string = value_string.lstrip('0')
    if len(value_string) == 0:
        return DIGITS['0']

    result_symbols = _get_value(value_string)
    # verbalized number starting with '一十*' is abbreviated as `十*`
    if len(result_symbols) >= 2 and result_symbols[0] == DIGITS[
            '1'] and result_symbols[1] == UNITS[1]:
        result_symbols = result_symbols[1:]
    return ''.join(result_symbols)


def verbalize_digit(value_string: str, alt_one=False) -> str:
    result_symbols = [DIGITS[digit] for digit in value_string]
    result = ''.join(result_symbols)
    if alt_one:
        result = result.replace("一", "幺")
    return result


def num2str(value_string: str) -> str:
    integer_decimal = value_string.split('.')
    if len(integer_decimal) == 1:
        integer = integer_decimal[0]
        decimal = ''
    elif len(integer_decimal) == 2:
        integer, decimal = integer_decimal
    else:
        raise ValueError(
            f"The value string: '${value_string}' has more than one point in it."
        )

    result = verbalize_cardinal(integer)

    decimal = decimal.rstrip('0')
    if decimal:
        # '.22' is verbalized as '零点二二'
        # '3.20' is verbalized as '三点二
        result = result if result else "零"
        result += '点' + verbalize_digit(decimal)
    return result



import os,sys,numpy as np
import traceback
from scipy.io import wavfile
# parent_directory = os.path.dirname(os.path.abspath(__file__))
# sys.path.append(parent_directory)
from tools.my_utils import load_audio
from slicer2 import Slicer

def slice(inp,opt_root,threshold,min_length,min_interval,hop_size,max_sil_kept,_max,alpha,i_part,all_part):
    os.makedirs(opt_root,exist_ok=True)
    if os.path.isfile(inp):
        input=[inp]
    elif os.path.isdir(inp):
        input=[os.path.join(inp, name) for name in sorted(list(os.listdir(inp)))]
    else:
        return "输入路径存在但既不是文件也不是文件夹"
    slicer = Slicer(
        sr=32000,  # 长音频采样率
        threshold=      int(threshold),  # 音量小于这个值视作静音的备选切割点
        min_length=     int(min_length),  # 每段最小多长，如果第一段太短一直和后面段连起来直到超过这个值
        min_interval=   int(min_interval),  # 最短切割间隔
        hop_size=       int(hop_size),  # 怎么算音量曲线，越小精度越大计算量越高（不是精度越大效果越好）
        max_sil_kept=   int(max_sil_kept),  # 切完后静音最多留多长
    )
    _max=float(_max)
    alpha=float(alpha)
    for inp_path in input[int(i_part)::int(all_part)]:
        # print(inp_path)
        try:
            name = os.path.basename(inp_path)
            audio = load_audio(inp_path, 32000)
            # print(audio.shape)
            for chunk, start, end in slicer.slice(audio):  # start和end是帧数
                tmp_max = np.abs(chunk).max()
                if(tmp_max>1):chunk/=tmp_max
                chunk = (chunk / tmp_max * (_max * alpha)) + (1 - alpha) * chunk
                wavfile.write(
                    "%s/%s_%010d_%010d.wav" % (opt_root, name, start, end),
                    32000,
                    # chunk.astype(np.float32),
                    (chunk * 32767).astype(np.int16),
                )
        except:
            print(inp_path,"->fail->",traceback.format_exc())
    return "执行完毕，请检查输出文件"

print(slice(*sys.argv[1:]))




from fastapi import APIRouter, File, UploadFile, Form, HTTPException, WebSocket, WebSocketDisconnect
from fastapi.responses import StreamingResponse
import io
import soundfile as sf
from gallama.data_classes import (
    TranscriptionResponse,
    TTSRequest,
    LanguageType
)

from typing import List, Literal, Optional, Dict
import asyncio
from ..dependencies import get_model_manager

# https://platform.openai.com/docs/api-reference/audio

router = APIRouter(prefix="/v1/audio", tags=["audio"])

# rest endpoint
@router.post("/transcriptions", response_model=TranscriptionResponse)
async def create_transcription(
    file: UploadFile = File(..., description="The audio file to transcribe."),
    model: str = Form(..., description="ID of the model to use."),
    language: Optional[LanguageType] = Form(None, description="The language of the input audio in ISO-639-1 format."),
    prompt: Optional[str] = Form(None, description="An optional text to guide the model's style or continue a previous segment."),
    response_format: Optional[Literal["json", "text", "srt", "verbose_json", "vtt"]] = Form("json", description="The format of the output."),
    temperature: Optional[float] = Form(0.0, description="The sampling temperature, between 0 and 1."),
    timestamp_granularities: Optional[List[Literal["word", "segment"]]] = Form(["segment"], description="The timestamp granularities for transcription.")
):
    """
    Transcribe an audio file into the input language.
    """
    model_manager = get_model_manager()
    stt = model_manager.get_model(model, _type="stt")
    # stt = model_manager.stt_dict.get(model, None)
    # if not stt:
    #     model_name, stt = next(iter(model_manager.stt_dict.items()))

    if stt is None:
        raise HTTPException(status_code=400, detail="Model not found")

    if response_format not in  ["json", "verbose_json"]:
        raise HTTPException(status_code=400, detail="Invalid response format, currently only json format supported")

    include_segments = False
    if response_format == "verbose_json" and "segment" in timestamp_granularities:
        include_segments = True

    transcribed_object = await stt.transcribe_async(
        audio=file.file,
        init_prompt=prompt,
        temperature=temperature,
        language=language,
        include_segments=include_segments,
        batch=True
    )

    return transcribed_object

@router.post("/speech")
async def create_speech(request: TTSRequest):
    # Validate the model, voice, and input length
    # if request.model not in SUPPORTED_MODELS:
    #     raise HTTPException(status_code=400, detail="Unsupported model")
    # if request.voice not in SUPPORTED_VOICES:
    #     raise HTTPException(status_code=400, detail="Unsupported voice")
    if len(request.input) > 4096:
        raise HTTPException(status_code=400, detail="Input text exceeds the maximum length of 4096 characters")

    global tts_dict
    model_manager = get_model_manager()
    tts = model_manager.get_model(request.model, _type="tts")

    # tts = model_manager.tts_dict[request.model]

    if len(request.input) > 4096:
        raise HTTPException(status_code=400, detail="Input text exceeds 4096 characters")

    if request.speed < 0.25 or request.speed > 4.0:
        raise HTTPException(status_code=400, detail="Speed must be between 0.25 and 4.0")

    # Get the current event loop at the start
    loop = asyncio.get_running_loop()

    # Generate audio asynchronously
    sampling_rate, audio_data = await tts.text_to_speech(
        text = request.input,
        stream = False,
        batching = True,
        batch_size = 3
    )

    # Convert to desired format using a thread pool
    buffer = io.BytesIO()
    if request.response_format not in ["mp3", "wav", "flac"]:
        raise HTTPException(status_code=400, detail=f"Unsupported format: {request.response_format}")

    # Write audio data to buffer in a thread pool
    await loop.run_in_executor(
        None,
        lambda: sf.write(buffer, audio_data, sampling_rate, format=request.response_format)
    )
    buffer.seek(0)

    # Set the appropriate media type
    media_types = {
        "mp3": "audio/mpeg",
        "wav": "audio/wav",
        "flac": "audio/flac"
    }

    return StreamingResponse(
        buffer,
        media_type=media_types[request.response_format],
        headers={
            "Content-Disposition": f'attachment; filename="speech.{request.response_format}"'
        }
    )



from typing import Optional, Deque, List, Union
import time
from PIL import Image
import io
from io import BytesIO
import base64
from collections import deque
from ..logger import logger


# Define the maximum total pixels for each setting
MAX_TOTAL_PIXELS = {
    "240p": 426 * 240,      # ~102K pixels
    "360p": 640 * 360,      # ~230K pixels
    "480p": 854 * 480,      # ~410K pixels
    "540p": 960 * 540,      # ~518K pixels
    "720p": 1280 * 720,     # ~921K pixels
    "900p": 1600 * 900,     # ~1.44M pixels
    "1080p": 1920 * 1080,   # ~2.07M pixels
}


class VideoFrame:
    def __init__(self, frame_data: bytes, timestamp: Optional[float] = None):
        """
        Initialize a VideoFrame object.

        Args:
            frame_data (bytes): The raw frame data as bytes.
            timestamp (Optional[float]): The timestamp of the frame. If not provided, the current time is used.
        """
        self.timestamp = timestamp if timestamp is not None else time.time()

        # Convert the frame data (bytes) into a PIL.Image object
        self.image = self._bytes_to_image(frame_data)

    def _bytes_to_image(self, frame_data: bytes) -> Image.Image:
        """
        Convert raw frame data (bytes) into a PIL.Image object.

        Args:
            frame_data (bytes): The raw frame data as bytes.

        Returns:
            Image.Image: The converted PIL.Image object.
        """
        try:
            # Use BytesIO to convert bytes into a file-like object
            image_stream = io.BytesIO(frame_data)
            image = Image.open(image_stream)
            return image
        except Exception as e:
            raise ValueError(f"Failed to convert frame data to PIL.Image: {e}")

    def get_image(self) -> Image.Image:
        """
        Get the PIL.Image object representing the frame.

        Returns:
            Image.Image: The PIL.Image object.
        """
        return self.image

    def get_timestamp(self) -> float:
        """
        Get the timestamp of the frame.

        Returns:
            float: The timestamp of the frame.
        """
        return self.timestamp


class VideoFrameCollection:
    def __init__(self, max_frames: int = 1000):
        self.max_frames = max_frames
        self.frames: Deque[VideoFrame] = deque(maxlen=max_frames)
        self.last_extraction_time: float = 0

    def add_frame(self, frame: VideoFrame, video_max_resolution: Optional[str] = None):
        """
        Add a frame to the collection. If the collection is full, the oldest frame is removed.
        The frame is rescaled based on total pixel count while maintaining aspect ratio.
        The width and height are adjusted to be multiples of 28.

        Args:
            frame (VideoFrame): The frame to add.
            video_max_resolution (Optional[str]): The maximum resolution setting (e.g., "720p").
        """
        if video_max_resolution and video_max_resolution in MAX_TOTAL_PIXELS:
            max_pixels = MAX_TOTAL_PIXELS[video_max_resolution]
            image = frame.get_image()
            width, height = image.size
            current_pixels = width * height

            if current_pixels > max_pixels:
                # Calculate the scaling ratio based on total pixels
                ratio = (max_pixels / current_pixels) ** 0.5  # Square root to apply evenly to both dimensions
                new_width = int(width * ratio)
                new_height = int(height * ratio)

                # Adjust the width and height to be multiples of 28
                new_width = (new_width // 28) * 28
                new_height = (new_height // 28) * 28

                # Ensure that the dimensions are at least 28
                new_width = max(new_width, 28)
                new_height = max(new_height, 28)

                logger.info(f"Rescaling to max resolution {video_max_resolution} "
                            f"(from {width}x{height} ({current_pixels:,} px) "
                            f"to {new_width}x{new_height} ({new_width * new_height:,} px))")

                resized_image = image.resize((new_width, new_height), Image.Resampling.LANCZOS)
                frame.image = resized_image

        self.frames.append(frame)

    def get_frames_between_timestamps(self, start_time: float, end_time: float) -> List[VideoFrame]:
        """
        Retrieve frames between the specified timestamps.
        Also update the extraction tracker to the latest extracted frame timestamp.
        """
        frames_in_range = [frame for frame in self.frames if start_time <= frame.timestamp <= end_time]
        if frames_in_range:
            # Update the tracker to the timestamp of the last extracted frame.
            self.last_extraction_time = frames_in_range[-1].timestamp
        return frames_in_range

    def get_all_frames(self, start_time: float = 0, only_since_last: bool = True) -> List[VideoFrame]:
        """
        Retrieve all frames in the collection.

        Args:
            only_since_last (bool): If True, returns only frames with a timestamp greater
                                    than the last extraction time. Otherwise, returns all frames.

        Also updates the tracker to the timestamp of the last extracted frame.
        """

        start_time_to_use = max(
            start_time,
            self.last_extraction_time if only_since_last else 0
        )
        logger.info(f"start time is {start_time_to_use}")

        if start_time_to_use:
            frames_to_return = [frame for frame in self.frames if frame.timestamp > start_time_to_use]
        else:
            frames_to_return = list(self.frames)

        if frames_to_return:
            # Update the tracker to the last frame's timestamp from the returned list.
            self.last_extraction_time = frames_to_return[-1].timestamp
        return frames_to_return

    def clear(self):
        """Clear all frames from the collection."""
        self.frames.clear()

    @classmethod
    def get_retained_frames_from_list(
        cls,
        frames: List['VideoFrame'],
        retained_video_frames_per_message: int,
        return_base64: bool = False,
    ) -> Union[List['VideoFrame'], List[str]]:
        """
        Select a number of frames from the provided list and return them.
        If return_base64 is True, converts each frame's image to a base64 Data URI
        (using PNG format), otherwise returns the VideoFrame objects directly.
        When only one frame is requested, the middle frame is selected.
        """
        if not frames:
            return "" if return_base64 else []

        total_frames = len(frames)
        if retained_video_frames_per_message >= total_frames:
            sampled_frames = frames
        else:
            if retained_video_frames_per_message == 1:
                # Select the middle frame
                middle_index = total_frames // 2
                sampled_frames = [frames[middle_index]]
            else:
                # Evenly sample indices from 0 to total_frames-1.
                indices = [
                    int(round(i * (total_frames - 1) / (retained_video_frames_per_message - 1)))
                    for i in range(retained_video_frames_per_message)
                ]
                sampled_frames = [frames[idx] for idx in indices]

        if not return_base64:
            return sampled_frames

        # Convert each sampled frame to a base64 encoded PNG Data URI

        return cls.convert_frames_to_base64(sampled_frames)

    @classmethod
    def sample_frames_time_based(cls, frames: List[VideoFrame], interval: int) -> List[VideoFrame]:
        """
        Retain one frame every `interval` seconds and always include the last frame.
        Assumes frames are in chronological order.
        """
        if not frames:
            return []
        selected = []
        # Start with the first frame
        last_selected_time = frames[0].timestamp
        selected.append(frames[0])
        for frame in frames[1:]:
            if frame.timestamp - last_selected_time >= interval:
                selected.append(frame)
                last_selected_time = frame.timestamp
        # Always include the last frame if not already selected
        if frames[-1] not in selected:
            selected.append(frames[-1])
        return selected

    @classmethod
    def convert_frames_to_base64(cls, frames: List['VideoFrame']) -> List[str]:
        """
        Convert a list of VideoFrame objects to a list of base64-encoded PNG Data URIs.

        Args:
            frames (List[VideoFrame]): The list of video frames to convert.

        Returns:
            List[str]: A list of Data URIs in the format "data:image/png;base64,<base64_data>".
        """
        base64_frames = []
        for frame in frames:
            buffered = BytesIO()
            frame.get_image().save(buffered, format="PNG")
            img_bytes = buffered.getvalue()
            b64_encoded = base64.b64encode(img_bytes).decode("utf-8")
            data_uri = f"data:image/png;base64,{b64_encoded}"
            base64_frames.append(data_uri)
        return base64_frames


# modified from https://github.com/lifeiteng/vall-e/blob/main/valle/modules/activation.py
from typing import Optional
from typing import Tuple
import torch
from torch import Tensor
from torch.nn import Linear
from torch.nn import Module
from torch.nn.init import constant_
from torch.nn.init import xavier_normal_
from torch.nn.init import xavier_uniform_
from torch.nn.modules.linear import NonDynamicallyQuantizableLinear
from torch.nn.parameter import Parameter

from torch.nn import functional as F
from AR.modules.patched_mha_with_cache import multi_head_attention_forward_patched

F.multi_head_attention_forward = multi_head_attention_forward_patched


class MultiheadAttention(Module):
    r"""Allows the model to jointly attend to information
    from different representation subspaces as described in the paper:
    `Attention Is All You Need <https://arxiv.org/abs/1706.03762>`_.

    Multi-Head Attention is defined as:

    .. math::
        \text{MultiHead}(Q, K, V) = \text{Concat}(head_1,\dots,head_h)W^O

    where :math:`head_i = \text{Attention}(QW_i^Q, KW_i^K, VW_i^V)`.

    ``forward()`` will use a special optimized implementation if all of the following
    conditions are met:

    - self attention is being computed (i.e., ``query``, ``key``, and ``value`` are the same tensor. This
      restriction will be loosened in the future.)
    - Either autograd is disabled (using ``torch.inference_mode`` or ``torch.no_grad``) or no tensor argument ``requires_grad``
    - training is disabled (using ``.eval()``)
    - dropout is 0
    - ``add_bias_kv`` is ``False``
    - ``add_zero_attn`` is ``False``
    - ``batch_first`` is ``True`` and the input is batched
    - ``kdim`` and ``vdim`` are equal to ``embed_dim``
    - at most one of ``key_padding_mask`` or ``attn_mask`` is passed
    - if a `NestedTensor <https://pytorch.org/docs/stable/nested.html>`_ is passed, neither ``key_padding_mask``
      nor ``attn_mask`` is passed

    If the optimized implementation is in use, a
    `NestedTensor <https://pytorch.org/docs/stable/nested.html>`_ can be passed for
    ``query``/``key``/``value`` to represent padding more efficiently than using a
    padding mask. In this case, a `NestedTensor <https://pytorch.org/docs/stable/nested.html>`_
    will be returned, and an additional speedup proportional to the fraction of the input
    that is padding can be expected.

    Args:
        embed_dim: Total dimension of the model.
        num_heads: Number of parallel attention heads. Note that ``embed_dim`` will be split
            across ``num_heads`` (i.e. each head will have dimension ``embed_dim // num_heads``).
        dropout: Dropout probability on ``attn_output_weights``. Default: ``0.0`` (no dropout).
        bias: If specified, adds bias to input / output projection layers. Default: ``True``.
        add_bias_kv: If specified, adds bias to the key and value sequences at dim=0. Default: ``False``.
        add_zero_attn: If specified, adds a new batch of zeros to the key and value sequences at dim=1.
            Default: ``False``.
        kdim: Total number of features for keys. Default: ``None`` (uses ``kdim=embed_dim``).
        vdim: Total number of features for values. Default: ``None`` (uses ``vdim=embed_dim``).
        batch_first: If ``True``, then the input and output tensors are provided
            as (batch, seq, feature). Default: ``False`` (seq, batch, feature).

    Examples::

        >>> # xdoctest: +SKIP
        >>> multihead_attn = nn.MultiheadAttention(embed_dim, num_heads)
        >>> attn_output, attn_output_weights = multihead_attn(query, key, value)

    """
    __constants__ = ["batch_first"]
    bias_k: Optional[torch.Tensor]
    bias_v: Optional[torch.Tensor]

    def __init__(
        self,
        embed_dim,
        num_heads,
        dropout=0.0,
        bias=True,
        add_bias_kv=False,
        add_zero_attn=False,
        kdim=None,
        vdim=None,
        batch_first=False,
        linear1_cls=Linear,
        linear2_cls=Linear,
        device=None,
        dtype=None,
    ) -> None:
        factory_kwargs = {"device": device, "dtype": dtype}
        super(MultiheadAttention, self).__init__()
        self.embed_dim = embed_dim
        self.kdim = kdim if kdim is not None else embed_dim
        self.vdim = vdim if vdim is not None else embed_dim
        self._qkv_same_embed_dim = self.kdim == embed_dim and self.vdim == embed_dim

        self.num_heads = num_heads
        self.dropout = dropout
        self.batch_first = batch_first
        self.head_dim = embed_dim // num_heads
        assert (
            self.head_dim * num_heads == self.embed_dim
        ), "embed_dim must be divisible by num_heads"

        if add_bias_kv:
            self.bias_k = Parameter(torch.empty((1, 1, embed_dim), **factory_kwargs))
            self.bias_v = Parameter(torch.empty((1, 1, embed_dim), **factory_kwargs))
        else:
            self.bias_k = self.bias_v = None

        if linear1_cls == Linear:
            if not self._qkv_same_embed_dim:
                self.q_proj_weight = Parameter(
                    torch.empty((embed_dim, embed_dim), **factory_kwargs)
                )
                self.k_proj_weight = Parameter(
                    torch.empty((embed_dim, self.kdim), **factory_kwargs)
                )
                self.v_proj_weight = Parameter(
                    torch.empty((embed_dim, self.vdim), **factory_kwargs)
                )
                self.register_parameter("in_proj_weight", None)
            else:
                self.in_proj_weight = Parameter(
                    torch.empty((3 * embed_dim, embed_dim), **factory_kwargs)
                )
                self.register_parameter("q_proj_weight", None)
                self.register_parameter("k_proj_weight", None)
                self.register_parameter("v_proj_weight", None)

            if bias:
                self.in_proj_bias = Parameter(
                    torch.empty(3 * embed_dim, **factory_kwargs)
                )
            else:
                self.register_parameter("in_proj_bias", None)
            self.out_proj = NonDynamicallyQuantizableLinear(
                embed_dim, embed_dim, bias=bias, **factory_kwargs
            )

            self._reset_parameters()
        else:
            if not self._qkv_same_embed_dim:
                raise NotImplementedError
            else:
                self.in_proj_linear = linear1_cls(
                    embed_dim, 3 * embed_dim, bias=bias, **factory_kwargs
                )
                self.in_proj_weight = self.in_proj_linear.weight

                self.register_parameter("q_proj_weight", None)
                self.register_parameter("k_proj_weight", None)
                self.register_parameter("v_proj_weight", None)

                if bias:
                    self.in_proj_bias = self.in_proj_linear.bias
                else:
                    self.register_parameter("in_proj_bias", None)

            self.out_proj = linear2_cls(
                embed_dim, embed_dim, bias=bias, **factory_kwargs
            )

            if self.bias_k is not None:
                xavier_normal_(self.bias_k)
            if self.bias_v is not None:
                xavier_normal_(self.bias_v)

        self.add_zero_attn = add_zero_attn

    def _reset_parameters(self):
        if self._qkv_same_embed_dim:
            xavier_uniform_(self.in_proj_weight)
        else:
            xavier_uniform_(self.q_proj_weight)
            xavier_uniform_(self.k_proj_weight)
            xavier_uniform_(self.v_proj_weight)

        if self.in_proj_bias is not None:
            constant_(self.in_proj_bias, 0.0)
            constant_(self.out_proj.bias, 0.0)

        if self.bias_k is not None:
            xavier_normal_(self.bias_k)
        if self.bias_v is not None:
            xavier_normal_(self.bias_v)

    def __setstate__(self, state):
        # Support loading old MultiheadAttention checkpoints generated by v1.1.0
        if "_qkv_same_embed_dim" not in state:
            state["_qkv_same_embed_dim"] = True

        super(MultiheadAttention, self).__setstate__(state)

    def forward(
        self,
        query: Tensor,
        key: Tensor,
        value: Tensor,
        key_padding_mask: Optional[Tensor] = None,
        need_weights: bool = True,
        attn_mask: Optional[Tensor] = None,
        average_attn_weights: bool = True,
        cache=None,
    ) -> Tuple[Tensor, Optional[Tensor]]:
        r"""
        Args:
            query: Query embeddings of shape :math:`(L, E_q)` for unbatched input, :math:`(L, N, E_q)` when ``batch_first=False``
                or :math:`(N, L, E_q)` when ``batch_first=True``, where :math:`L` is the target sequence length,
                :math:`N` is the batch size, and :math:`E_q` is the query embedding dimension ``embed_dim``.
                Queries are compared against key-value pairs to produce the output.
                See "Attention Is All You Need" for more details.
            key: Key embeddings of shape :math:`(S, E_k)` for unbatched input, :math:`(S, N, E_k)` when ``batch_first=False``
                or :math:`(N, S, E_k)` when ``batch_first=True``, where :math:`S` is the source sequence length,
                :math:`N` is the batch size, and :math:`E_k` is the key embedding dimension ``kdim``.
                See "Attention Is All You Need" for more details.
            value: Value embeddings of shape :math:`(S, E_v)` for unbatched input, :math:`(S, N, E_v)` when
                ``batch_first=False`` or :math:`(N, S, E_v)` when ``batch_first=True``, where :math:`S` is the source
                sequence length, :math:`N` is the batch size, and :math:`E_v` is the value embedding dimension ``vdim``.
                See "Attention Is All You Need" for more details.
            key_padding_mask: If specified, a mask of shape :math:`(N, S)` indicating which elements within ``key``
                to ignore for the purpose of attention (i.e. treat as "padding"). For unbatched `query`, shape should be :math:`(S)`.
                Binary and byte masks are supported.
                For a binary mask, a ``True`` value indicates that the corresponding ``key`` value will be ignored for
                the purpose of attention. For a float mask, it will be directly added to the corresponding ``key`` value.
            need_weights: If specified, returns ``attn_output_weights`` in addition to ``attn_outputs``.
                Default: ``True``.
            attn_mask: If specified, a 2D or 3D mask preventing attention to certain positions. Must be of shape
                :math:`(L, S)` or :math:`(N\cdot\text{num\_heads}, L, S)`, where :math:`N` is the batch size,
                :math:`L` is the target sequence length, and :math:`S` is the source sequence length. A 2D mask will be
                broadcasted across the batch while a 3D mask allows for a different mask for each entry in the batch.
                Binary, byte, and float masks are supported. For a binary mask, a ``True`` value indicates that the
                corresponding position is not allowed to attend. For a byte mask, a non-zero value indicates that the
                corresponding position is not allowed to attend. For a float mask, the mask values will be added to
                the attention weight.
            average_attn_weights: If true, indicates that the returned ``attn_weights`` should be averaged across
                heads. Otherwise, ``attn_weights`` are provided separately per head. Note that this flag only has an
                effect when ``need_weights=True``. Default: ``True`` (i.e. average weights across heads)

        Outputs:
            - **attn_output** - Attention outputs of shape :math:`(L, E)` when input is unbatched,
              :math:`(L, N, E)` when ``batch_first=False`` or :math:`(N, L, E)` when ``batch_first=True``,
              where :math:`L` is the target sequence length, :math:`N` is the batch size, and :math:`E` is the
              embedding dimension ``embed_dim``.
            - **attn_output_weights** - Only returned when ``need_weights=True``. If ``average_attn_weights=True``,
              returns attention weights averaged across heads of shape :math:`(L, S)` when input is unbatched or
              :math:`(N, L, S)`, where :math:`N` is the batch size, :math:`L` is the target sequence length, and
              :math:`S` is the source sequence length. If ``average_attn_weights=False``, returns attention weights per
              head of shape :math:`(\text{num\_heads}, L, S)` when input is unbatched or :math:`(N, \text{num\_heads}, L, S)`.

            .. note::
                `batch_first` argument is ignored for unbatched inputs.
        """
        is_batched = query.dim() == 3
        if key_padding_mask is not None:
            _kpm_dtype = key_padding_mask.dtype
            if _kpm_dtype != torch.bool and not torch.is_floating_point(
                key_padding_mask
            ):
                raise AssertionError(
                    "only bool and floating types of key_padding_mask are supported"
                )
        why_not_fast_path = ""
        if not is_batched:
            why_not_fast_path = (
                f"input not batched; expected query.dim() of 3 but got {query.dim()}"
            )
        elif query is not key or key is not value:
            # When lifting this restriction, don't forget to either
            # enforce that the dtypes all match or test cases where
            # they don't!
            why_not_fast_path = "non-self attention was used (query, key, and value are not the same Tensor)"
        elif self.in_proj_bias is not None and query.dtype != self.in_proj_bias.dtype:
            why_not_fast_path = f"dtypes of query ({query.dtype}) and self.in_proj_bias ({self.in_proj_bias.dtype}) don't match"
        elif (
            self.in_proj_weight is not None and query.dtype != self.in_proj_weight.dtype
        ):
            # this case will fail anyway, but at least they'll get a useful error message.
            why_not_fast_path = f"dtypes of query ({query.dtype}) and self.in_proj_weight ({self.in_proj_weight.dtype}) don't match"
        elif self.training:
            why_not_fast_path = "training is enabled"
        elif not self.batch_first:
            why_not_fast_path = "batch_first was not True"
        elif self.bias_k is not None:
            why_not_fast_path = "self.bias_k was not None"
        elif self.bias_v is not None:
            why_not_fast_path = "self.bias_v was not None"
        elif self.dropout:
            why_not_fast_path = f"dropout was {self.dropout}, required zero"
        elif self.add_zero_attn:
            why_not_fast_path = "add_zero_attn was enabled"
        elif not self._qkv_same_embed_dim:
            why_not_fast_path = "_qkv_same_embed_dim was not True"
        elif attn_mask is not None:
            why_not_fast_path = "attn_mask was not None"
        elif query.is_nested and key_padding_mask is not None:
            why_not_fast_path = (
                "key_padding_mask is not supported with NestedTensor input"
            )
        elif self.num_heads % 2 == 1:
            why_not_fast_path = "num_heads is odd"
        elif torch.is_autocast_enabled():
            why_not_fast_path = "autocast is enabled"

        if not why_not_fast_path:
            tensor_args = (
                query,
                key,
                value,
                self.in_proj_weight,
                self.in_proj_bias,
                self.out_proj.weight,
                self.out_proj.bias,
            )
            # We have to use list comprehensions below because TorchScript does not support
            # generator expressions.
            if torch.overrides.has_torch_function(tensor_args):
                why_not_fast_path = "some Tensor argument has_torch_function"
            elif not all(
                [
                    (x is None or x.is_cuda or "cpu" in str(x.device))
                    for x in tensor_args
                ]
            ):
                why_not_fast_path = "some Tensor argument is neither CUDA nor CPU"
            elif torch.is_grad_enabled() and any(
                [x is not None and x.requires_grad for x in tensor_args]
            ):
                why_not_fast_path = (
                    "grad is enabled and at least one of query or the "
                    "input/output projection weights or biases requires_grad"
                )
            if not why_not_fast_path:
                return torch._native_multi_head_attention(
                    query,
                    key,
                    value,
                    self.embed_dim,
                    self.num_heads,
                    self.in_proj_weight,
                    self.in_proj_bias,
                    self.out_proj.weight,
                    self.out_proj.bias,
                    key_padding_mask if key_padding_mask is not None else attn_mask,
                    need_weights,
                    average_attn_weights,
                    1
                    if key_padding_mask is not None
                    else 0
                    if attn_mask is not None
                    else None,
                )

        any_nested = query.is_nested or key.is_nested or value.is_nested
        assert not any_nested, (
            "MultiheadAttention does not support NestedTensor outside of its fast path. "
            + f"The fast path was not hit because {why_not_fast_path}"
        )

        if self.batch_first and is_batched:
            # make sure that the transpose op does not affect the "is" property
            if key is value:
                if query is key:
                    query = key = value = query.transpose(1, 0)
                else:
                    query, key = [x.transpose(1, 0) for x in (query, key)]
                    value = key
            else:
                query, key, value = [x.transpose(1, 0) for x in (query, key, value)]

        if not self._qkv_same_embed_dim:
            attn_output, attn_output_weights = F.multi_head_attention_forward(
                query,
                key,
                value,
                self.embed_dim,
                self.num_heads,
                self.in_proj_weight,
                self.in_proj_bias,
                self.bias_k,
                self.bias_v,
                self.add_zero_attn,
                self.dropout,
                self.out_proj.weight,
                self.out_proj.bias,
                training=self.training,
                key_padding_mask=key_padding_mask,
                need_weights=need_weights,
                attn_mask=attn_mask,
                use_separate_proj_weight=True,
                q_proj_weight=self.q_proj_weight,
                k_proj_weight=self.k_proj_weight,
                v_proj_weight=self.v_proj_weight,
                average_attn_weights=average_attn_weights,
                cache=cache,
            )
        else:
            attn_output, attn_output_weights = F.multi_head_attention_forward(
                query,
                key,
                value,
                self.embed_dim,
                self.num_heads,
                self.in_proj_weight,
                self.in_proj_bias,
                self.bias_k,
                self.bias_v,
                self.add_zero_attn,
                self.dropout,
                self.out_proj.weight,
                self.out_proj.bias,
                training=self.training,
                key_padding_mask=key_padding_mask,
                need_weights=need_weights,
                attn_mask=attn_mask,
                average_attn_weights=average_attn_weights,
                cache=cache,
            )
        if self.batch_first and is_batched:
            return attn_output.transpose(1, 0), attn_output_weights
        else:
            return attn_output, attn_output_weights



from pydantic import BaseModel, Field, validator, ConfigDict, RootModel, field_validator, constr, model_validator
from typing import Optional, Literal, List, Dict, Union, Any, Type


class ModelRequest(BaseModel):
    model_id: str

    class Config:
        # Override protected namespaces to suppress warning
        protected_namespaces = ()


class ModelInstanceInfo(BaseModel):
    model_name: str
    port: int
    pid: int  # Changed from process to pid
    status: str
    model_type: Literal["stt", "llm", "tts", "embedding"]
    strict: bool = Field(description="whether require api call to match model name or not", default=False)

    class Config:
        # Override protected namespaces to suppress warning
        protected_namespaces = ()


class ModelInfo(BaseModel):
    instances: List[ModelInstanceInfo]


class AgentWithThinking(BaseModel):
    model: str = Field(description="The name of the model")
    thinking_template: Optional[str] = Field(default=None, description="The XML thinking for the agent")


class MixtureOfAgents(BaseModel):
    agent_list: List[Union[str, AgentWithThinking]]

class StopModelByPort(BaseModel):
    port: int


import json
import os
import pathlib

default_param = {}
default_param["bins"] = 768
default_param["unstable_bins"] = 9  # training only
default_param["reduction_bins"] = 762  # training only
default_param["sr"] = 44100
default_param["pre_filter_start"] = 757
default_param["pre_filter_stop"] = 768
default_param["band"] = {}


default_param["band"][1] = {
    "sr": 11025,
    "hl": 128,
    "n_fft": 960,
    "crop_start": 0,
    "crop_stop": 245,
    "lpf_start": 61,  # inference only
    "res_type": "polyphase",
}

default_param["band"][2] = {
    "sr": 44100,
    "hl": 512,
    "n_fft": 1536,
    "crop_start": 24,
    "crop_stop": 547,
    "hpf_start": 81,  # inference only
    "res_type": "sinc_best",
}


def int_keys(d):
    r = {}
    for k, v in d:
        if k.isdigit():
            k = int(k)
        r[k] = v
    return r


class ModelParameters(object):
    def __init__(self, config_path=""):
        if ".pth" == pathlib.Path(config_path).suffix:
            import zipfile

            with zipfile.ZipFile(config_path, "r") as zip:
                self.param = json.loads(
                    zip.read("param.json"), object_pairs_hook=int_keys
                )
        elif ".json" == pathlib.Path(config_path).suffix:
            with open(config_path, "r") as f:
                self.param = json.loads(f.read(), object_pairs_hook=int_keys)
        else:
            self.param = default_param

        for k in [
            "mid_side",
            "mid_side_b",
            "mid_side_b2",
            "stereo_w",
            "stereo_n",
            "reverse",
        ]:
            if not k in self.param:
                self.param[k] = False



from silero_vad import load_silero_vad, VADIterator
from ..data_classes.realtime_client_proto import TurnDetectionConfig
from ..data_classes.internal_ws import SpeechState
import numpy as np
from typing import Tuple, Optional, Dict, Literal
import torch
from dataclasses import dataclass
import samplerate
from collections import deque
from .audio_preprocessor import AudioPreprocessor
import os
import soundfile as sf
from datetime import datetime

from ..dependencies_server import get_server_logger

logger = get_server_logger()



@dataclass
class VADState:
    is_speech: bool = False
    speech_start_time: float = 0.0
    accumulated_silence: float = 0.0
    last_window_time: float = 0.0
    potential_speech_start: float = 0.0
    speech_duration: float = 0.0
    speech_samples: Optional[np.ndarray] = None
    original_speech_samples: Optional[np.ndarray] = None  # Store original sample rate speech
    current_state: SpeechState = "no_speech"


@dataclass
class PreprocessingConfig:
    """Separate configuration for audio preprocessing"""
    enable_highpass: bool = True
    enable_compression: bool = True
    highpass_cutoff: float = 100
    compression_threshold: float = -20
    compression_ratio: float = 3.0
    attack_time: float = 0.005
    release_time: float = 0.1
    makeup_gain: float = 7.0


class VADProcessor:
    def __init__(self, turn_detection_config: TurnDetectionConfig, preprocessing_config: PreprocessingConfig = None,
                 input_sample_rate: int = 24000, debug: bool = False,
                 # debug_folder_path: str = "/home/remichu/work/ML/gallama/experiment/log"):
                 debug_folder_path: str = ""):
        self.model = load_silero_vad()
        self.model.eval()



        # Sample rate configuration
        self.input_sample_rate = input_sample_rate

        # minimum audio chunk sample:
        self.min_audio_chunk_sample = int(self.input_sample_rate / 10)

        self.vad_sample_rate = 16000
        self.resampler = samplerate.Resampler('sinc_best', channels=1)
        self.resample_ratio = self.vad_sample_rate / self.input_sample_rate

        # Debug settings
        self.debug = debug
        self.debug_folder_path = debug_folder_path
        if self.debug and not os.path.exists(debug_folder_path):
            os.makedirs(debug_folder_path)

        self.threshold = turn_detection_config.threshold
        self.silence_duration_ms = turn_detection_config.silence_duration_ms
        self.min_speech_duration_ms = 250
        self.create_response = turn_detection_config.create_response

        # Fixed parameters
        self.window_size_samples = 512
        self.speech_pad_ms = 30

        # State management
        self.state = VADState()
        self.current_chunk = np.array([], dtype=np.float32)

        # Initialize VAD iterator
        self.vad_iterator = VADIterator(
            model=self.model,
            threshold=self.threshold,
            sampling_rate=self.vad_sample_rate,
            min_silence_duration_ms=self.silence_duration_ms,
            speech_pad_ms=self.speech_pad_ms
        )
        self.total_audio_processed = 0.0

    def _resample_audio(self, audio: np.ndarray) -> np.ndarray:
        """Resample audio from input sample rate to VAD sample rate"""
        if audio.size == 0:
            return np.array([], dtype=np.float32)
        return self.resampler.process(audio, self.resample_ratio)

    def _save_debug_audio(self, audio_data: np.ndarray):
        """Save debug audio to file"""
        if not self.debug or not self.debug_folder_path:
            return

        timestamp = datetime.now().strftime("%Y%m%d_%H%M%S_%f")
        filename = os.path.join(self.debug_folder_path, f"speech_segment_{timestamp}.wav")

        # Save using input sample rate for original quality
        sf.write(filename, audio_data, self.input_sample_rate)
        logger.debug(f"Saved debug audio to {filename}")

    def process_audio_chunk(self, audio_float: np.ndarray) -> Tuple[SpeechState, Optional[Dict[str, float]]]:
        """
        Process an audio chunk and determine its speech state
        Args:
            audio_float: Numpy array of float32 audio data, normalized to [-1.0, 1.0]
        Returns: (speech_state, speech_event)
            speech_state: Current state of speech detection
            speech_event: Optional dict with 'start' or 'end' timestamps, where start represents
                         the beginning of the potential_speech buffer that was confirmed as speech
        """
        if self.audio_preprocessor:
            audio_float = self.audio_preprocessor.process_float_chunk(audio_float)

        if audio_float.size == 0:
            return self.state.current_state, None

        # Store original audio for speech segments if debugging
        if self.state.current_state in ["is_speaking", "potential_speech"]:
            if self.state.original_speech_samples is None:
                self.state.original_speech_samples = audio_float
            else:
                self.state.original_speech_samples = np.concatenate(
                    [self.state.original_speech_samples, audio_float])

        # Resample audio to VAD sample rate
        resampled_audio = self._resample_audio(audio_float)
        self.current_chunk = np.concatenate([self.current_chunk, resampled_audio])

        latest_event = None
        prev_state = self.state.current_state

        # Process complete windows
        while self.current_chunk.size >= self.window_size_samples:
            window = self.current_chunk[:self.window_size_samples]
            self.current_chunk = self.current_chunk[self.window_size_samples:]

            window_tensor = torch.from_numpy(window)
            speech_prob = self.model(window_tensor, self.vad_sample_rate).item()

            window_time = self.window_size_samples / self.vad_sample_rate
            self.total_audio_processed += window_time

            is_speech = speech_prob >= self.threshold

            # Update state and get events
            event = self._update_speech_state(is_speech)
            if event is not None:
                latest_event = event

        return self.state.current_state, latest_event

    def _update_speech_state(self, is_speech: bool) -> Optional[Dict[str, float]]:
        """Update speech state and return speech events"""
        event = None
        prev_state = self.state.current_state

        if is_speech:
            if self.state.current_state == "no_speech":
                self.state.current_state = "potential_speech"
                self.state.potential_speech_start = self.total_audio_processed

            elif self.state.current_state == "potential_speech":
                current_duration = (self.total_audio_processed - self.state.potential_speech_start) * 1000
                if current_duration >= self.min_speech_duration_ms:
                    self.state.current_state = "is_speaking"
                    # Use the potential_speech_start time for the speech start event
                    event = {'start': self.state.potential_speech_start}
                    logger.debug(f"Speech started at {self.state.potential_speech_start:.3f}s after reaching minimum duration: {current_duration:.1f}ms")

            self.state.accumulated_silence = 0

        else:  # not is_speech
            if self.state.current_state == "potential_speech":
                # If we were in potential speech but got silence, go back to no speech
                self.state.current_state = "no_speech"
                self.state.potential_speech_start = 0.0
                self.state.original_speech_samples = None

            elif self.state.current_state == "is_speaking":
                self.state.accumulated_silence += self.window_size_samples / self.vad_sample_rate

                if self.state.accumulated_silence >= (self.silence_duration_ms / 1000):
                    self.state.current_state = "speak_end"
                    event = {'end': self.total_audio_processed}

                    if self.debug and self.state.original_speech_samples is not None:
                        self._save_debug_audio(self.state.original_speech_samples)

                    # Reset state after speech end
                    self.state.is_speech = False
                    self.state.accumulated_silence = 0
                    self.state.potential_speech_start = 0.0
                    self.state.speech_duration = 0.0
                    self.state.original_speech_samples = None
                    self.total_audio_processed = 0.0

                    # After one frame in speak_end, go back to no_speech
                    self.state.current_state = "no_speech"

        return event


    def reset(self):
        """Reset all state"""
        if self.audio_preprocessor:
            self.audio_preprocessor.reset()
        self.vad_iterator.reset_states()
        self.current_chunk = np.array([], dtype=np.float32)
        self.state = VADState()
        self.total_audio_processed = 0.0




import platform,os,traceback
import ffmpeg
import numpy as np
import gradio as gr
from tools.i18n.i18n import I18nAuto
import pandas as pd
i18n = I18nAuto(language=os.environ.get('language','Auto'))

def load_audio(file, sr):
    try:
        # https://github.com/openai/whisper/blob/main/whisper/audio.py#L26
        # This launches a subprocess to decode audio while down-mixing and resampling as necessary.
        # Requires the ffmpeg CLI and `ffmpeg-python` package to be installed.
        file = clean_path(file)  # 防止小白拷路径头尾带了空格和"和回车
        if os.path.exists(file) == False:
            raise RuntimeError(
                "You input a wrong audio path that does not exists, please fix it!"
            )
        out, _ = (
            ffmpeg.input(file, threads=0)
            .output("-", format="f32le", acodec="pcm_f32le", ac=1, ar=sr)
            .run(cmd=["ffmpeg", "-nostdin"], capture_stdout=True, capture_stderr=True)
        )
    except Exception as e:
        traceback.print_exc()
        raise RuntimeError(i18n("音频加载失败"))

    return np.frombuffer(out, np.float32).flatten()


def clean_path(path_str:str):
    if path_str.endswith(('\\','/')):
        return clean_path(path_str[0:-1])
    path_str = path_str.replace('/', os.sep).replace('\\', os.sep)
    return path_str.strip(" ").strip('\'').strip("\n").strip('"').strip(" ").strip("\u202a")


def check_for_existance(file_list:list=None,is_train=False,is_dataset_processing=False):
    files_status=[]
    if is_train == True and file_list:
        file_list.append(os.path.join(file_list[0],'2-name2text.txt'))
        file_list.append(os.path.join(file_list[0],'3-bert'))
        file_list.append(os.path.join(file_list[0],'4-cnhubert'))
        file_list.append(os.path.join(file_list[0],'5-wav32k'))
        file_list.append(os.path.join(file_list[0],'6-name2semantic.tsv'))
    for file in file_list:
        if os.path.exists(file):files_status.append(True)
        else:files_status.append(False)
    if sum(files_status)!=len(files_status):
        if is_train:
            for file,status in zip(file_list,files_status):
                if status:pass
                else:gr.Warning(file)
            gr.Warning(i18n('以下文件或文件夹不存在'))
            return False
        elif is_dataset_processing:
            if files_status[0]:
                return True
            elif not files_status[0]:
                gr.Warning(file_list[0])
            elif not files_status[1] and file_list[1]:
                gr.Warning(file_list[1])
            gr.Warning(i18n('以下文件或文件夹不存在'))
            return False
        else:
            if file_list[0]:
                gr.Warning(file_list[0])
                gr.Warning(i18n('以下文件或文件夹不存在'))
            else:
                gr.Warning(i18n('路径不能为空'))
            return False
    return True

def check_details(path_list=None,is_train=False,is_dataset_processing=False):
    if is_dataset_processing:
        list_path, audio_path = path_list
        if (not list_path.endswith('.list')):
            gr.Warning(i18n('请填入正确的List路径'))
            return
        if audio_path:
            if not os.path.isdir(audio_path):
                gr.Warning(i18n('请填入正确的音频文件夹路径'))
                return
        with open(list_path,"r",encoding="utf8")as f:
            line=f.readline().strip("\n").split("\n")
        wav_name, _, __, ___ = line[0].split("|")
        wav_name=clean_path(wav_name)
        if (audio_path != "" and audio_path != None):
            wav_name = os.path.basename(wav_name)
            wav_path = "%s/%s"%(audio_path, wav_name)
        else:
            wav_path=wav_name
        if os.path.exists(wav_path):
            ...
        else:
            gr.Warning(i18n('路径错误'))
        return
    if is_train:
        path_list.append(os.path.join(path_list[0],'2-name2text.txt'))
        path_list.append(os.path.join(path_list[0],'4-cnhubert'))
        path_list.append(os.path.join(path_list[0],'5-wav32k'))
        path_list.append(os.path.join(path_list[0],'6-name2semantic.tsv'))
        phone_path, hubert_path, wav_path, semantic_path = path_list[1:]
        with open(phone_path,'r',encoding='utf-8') as f:
            if f.read(1):...
            else:gr.Warning(i18n('缺少音素数据集'))
        if os.listdir(hubert_path):...
        else:gr.Warning(i18n('缺少Hubert数据集'))
        if os.listdir(wav_path):...
        else:gr.Warning(i18n('缺少音频数据集'))
        df = pd.read_csv(
            semantic_path, delimiter="\t", encoding="utf-8"
        )
        if len(df) >= 1:...
        else:gr.Warning(i18n('缺少语义数据集'))



import asyncio
import json
from typing import Optional, Dict, Any

import websockets
from pydantic import BaseModel
from websockets import WebSocketException
from websockets.protocol import State

from gallama.logger import logger


class WebSocketClient:
    def __init__(
            self,
            uri: str,
            reconnect_interval: float = 5.0,
            max_retries: int = 5,
            auto_reconnect: bool = True,
            max_receive_retries: int = 3  # Added parameter for receive_message retries
    ):
        self.uri = uri
        self.connection: Optional[websockets.WebSocketClientProtocol] = None
        self.reconnect_interval = reconnect_interval
        self.max_retries = max_retries
        self.auto_reconnect = auto_reconnect
        self.max_receive_retries = max_receive_retries  # Store the max receive retries
        self.retry_count = 0
        self.is_connecting = False

        self.websocket_options = {
            "ping_interval": 20,  # Enable periodic ping to detect connection issues
            "ping_timeout": 20,
            "close_timeout": 300,
            "max_size": None,
            "open_timeout": 60
        }

    async def _wait_for_connection(self, timeout: float = 30.0):
        """Wait for an ongoing connection attempt to complete."""
        start_time = asyncio.get_event_loop().time()
        while self.is_connecting:
            if asyncio.get_event_loop().time() - start_time > timeout:
                logger.warning("Timeout waiting for connection attempt")
                return False
            await asyncio.sleep(0.1)

        # After waiting, check if we actually have a valid connection
        return self.connection is not None and self.connection.state == State.OPEN

    async def ensure_connection(self) -> bool:
        """Ensures that the connection is active and attempts to reconnect if necessary."""
        try:
            # First check if we have an active connection
            if self.connection and self.connection.state == State.OPEN:
                try:
                    # Verify connection is truly alive with ping
                    pong_waiter = await self.connection.ping()
                    await asyncio.wait_for(pong_waiter, timeout=10)
                    return True
                except Exception as e:
                    logger.debug(f"Ping check failed: {str(e)}")
                    # Connection is not truly alive
                    self.connection = None

            # If another task is already trying to connect, wait for it
            if self.is_connecting:
                logger.debug("Another task is connecting, waiting...")
                return await self._wait_for_connection()

            # Set connecting flag before attempting connection
            self.is_connecting = True
            try:
                return await self.connect()
            finally:
                self.is_connecting = False

        except Exception as e:
            logger.debug(f"Connection check failed: {str(e)}")
            return False

    async def connect(self) -> bool:
        """Establishes a WebSocket connection with retry logic."""
        if self.is_connecting:
            return await self._wait_for_connection()

        self.is_connecting = True
        try:
            while self.retry_count < self.max_retries:
                try:
                    self.connection = await websockets.connect(self.uri, **self.websocket_options)
                    self.retry_count = 0
                    logger.info("Successfully connected to WebSocket server")
                    return True
                except WebSocketException as e:
                    self.retry_count += 1
                    logger.warning(f"Connection attempt {self.retry_count} failed: {str(e)}")

                    if self.retry_count >= self.max_retries:
                        logger.error("Max retry attempts reached")
                        return False

                    await asyncio.sleep(self.reconnect_interval)
            return False
        finally:
            self.is_connecting = False

    async def send_message(self, message: Dict[str, Any]) -> bool:
        """Sends a dictionary message, handling reconnection if necessary."""
        try:
            if not await self.ensure_connection():
                return False

            await self.connection.send(json.dumps(message))
            return True
        except WebSocketException as e:
            logger.error(f"Error sending message: {str(e)}")
            if self.auto_reconnect:
                self.connection = None
                return await self.send_message(message)
            return False

    async def send_pydantic_message(self, message: BaseModel) -> bool:
        """Sends a Pydantic model message, handling reconnection if necessary."""
        try:
            if not await self.ensure_connection():
                return False

            await self.connection.send(message.model_dump_json())
            return True
        except WebSocketException as e:
            logger.error(f"Error sending Pydantic message: {str(e)}")
            if self.auto_reconnect:
                self.connection = None
                return await self.send_pydantic_message(message)
            return False

    async def receive_message(self, timeout: float = 30, max_retries: Optional[int] = None, disable_warning: bool = False) -> Optional[str]:
        """
        Receives a message from the WebSocket connection with retry logic.

        Args:
            timeout: Time in seconds to wait for each receive attempt
            max_retries: Maximum number of retry attempts (defaults to self.max_receive_retries)

        Returns:
            The received message as a string, or None if receiving failed
        """
        retries = 0
        max_attempts = max_retries if max_retries is not None else self.max_receive_retries

        while retries <= max_attempts:
            try:
                if not await self.ensure_connection():
                    logger.warning(f"Connection failed on receive attempt {retries + 1}")
                    retries += 1
                    if retries > max_attempts:
                        return None
                    continue

                return await asyncio.wait_for(self.connection.recv(), timeout=timeout)

            except asyncio.TimeoutError:
                if not disable_warning:
                    logger.warning(f"Timeout on receive attempt {retries + 1}")

                retries += 1
                if retries > max_attempts:
                    if not disable_warning:
                        logger.error("Max receive retry attempts reached")
                    raise asyncio.TimeoutError

            except WebSocketException as e:
                logger.error(f"Error receiving message on attempt {retries + 1}: {str(e)}")
                if self.auto_reconnect:
                    self.connection = None
                    retries += 1
                    if retries > max_attempts:
                        logger.error("Max receive retry attempts reached")
                        return None
                    await asyncio.sleep(self.reconnect_interval)
                else:
                    return None

            except Exception as e:
                logger.error(f"Unexpected error on receive attempt {retries + 1}: {str(e)}")
                retries += 1
                if retries > max_attempts:
                    return None

    async def close(self):
        """Closes the WebSocket connection."""
        if self.connection:
            try:
                await self.connection.close()
            except WebSocketException as e:
                logger.error(f"Error closing connection: {str(e)}")
            finally:
                self.connection = None
                self.retry_count = 0

    async def __aenter__(self):
        """Async context manager entry."""
        await self.connect()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """Async context manager exit."""
        await self.close()


import uuid
import yaml
import base64
import struct
from lxml import etree
from pathlib import Path
from PIL import Image
import requests
import os
from io import BytesIO
from fastapi import Request
import json
from typing import Union, Optional, Tuple
from gallama.logger import logger

# Lazy import for Exllama
try:
    from exllamav2 import ExLlamaV2Tokenizer
except:
    ExLlamaV2Tokenizer = None

# Lazy import for llama_cpp
llama_cpp = None

def import_llama_cpp():
    global llama_cpp
    if llama_cpp is None:
        try:
            import llama_cpp
        except ImportError:
            raise ImportError("llama_cpp is not installed. Please install it to use Llama models.")
    return llama_cpp

def get_response_uid():
    return "cmpl-" + str(uuid.uuid4().hex)


def get_response_tool_uid():
    return "call_" + str(uuid.uuid4().hex)


def get_token_length(tokenizer, text) -> int:

    str_text = text
    if not isinstance(text, str):
        str_text = str(text)

    if ExLlamaV2Tokenizer and isinstance(tokenizer, ExLlamaV2Tokenizer):
        return tokenizer.num_tokens(str_text)
    elif type(tokenizer).__name__ == 'Llama':  # Check the type name instead of using isinstance
        # this is because llama_cpp is optional dependency
        try:
            return len(tokenizer.tokenize(str_text.encode("utf-8"), add_bos=False))
        except AttributeError:
            # If tokenize method is not available, fallback to a general method
            return len(tokenizer.encode(str_text))
    else:
        return len(tokenizer.encode(str_text))

def read_yaml(file_path):
    try:
        with open(file_path, 'r') as file:
            data = yaml.safe_load(file)
            return data
    except FileNotFoundError:
        print(f"The file {file_path} does not exist.")
        return None
    except yaml.YAMLError as exc:
        print(f"Error reading YAML file: {exc}")
        return None


def floats_to_base64(float_array):
    """" convert list of floats to base64 for embedding response """

    # Pack the float array into binary format
    binary_data = struct.pack('f' * len(float_array), *float_array)

    # Encode the binary data to base64
    base64_encoded = base64.b64encode(binary_data)

    # Convert bytes to string for easy handling
    base64_string = base64_encoded.decode('utf-8')

    return base64_string


def parse_xml_to_dict(xml_string):
    # Parse the XML string
    root = etree.fromstring(xml_string)

    # Get the root tag
    root_tag = root.tag

    # Function to convert XML element to a nested dictionary
    def xml_to_dict(element):
        children = list(element)
        if not children:
            return element.text
        result = {}
        for child in children:
            child_dict = xml_to_dict(child)
            if child.tag in result:
                if not isinstance(result[child.tag], list):
                    result[child.tag] = [result[child.tag]]
                result[child.tag].append(child_dict)
            else:
                result[child.tag] = child_dict
        return result

    # Convert the XML to a nested dictionary
    xml_dict = {root.tag: xml_to_dict(root)}

    return xml_dict


def get_package_file_path(file_name: str) -> str:
    """
    Get the absolute path of a file within the gallama package.

    Args:
    file_name (str): The name of the file (e.g., 'app.py')

    Returns:
    str: The absolute path to the file
    """
    current_dir = Path(__file__).parent.parent
    file_path = current_dir / file_name
    if not file_path.exists():
        raise FileNotFoundError(f"File {file_name} not found in the gallama package.")
    return str(file_path.resolve())


# Util function to get a PIL image from a URL or from a file in the script's directory
def get_image(
    file: str = None,
    url: str =None,     # url of the image or base64 url in this format f"data:image/jpeg;base64,{base64_image}"
):
    assert (file or url) and not (file and url)

    if file:
        script_dir = os.path.dirname(os.path.abspath(__file__))
        file_path = os.path.join(script_dir, file)
        return Image.open(file_path)

    elif url:
        # Check if url starts with 'data:image' to handle base64 images
        if url.startswith("data:image"):
            # Extract the base64 data from the url string
            base64_data = url.split(",")[1]
            # Decode the base64 string
            image_data = base64.b64decode(base64_data)
            # Open the image from the decoded data
            return Image.open(BytesIO(image_data))
        else:
            # Assume url is a regular URL and fetch it as a stream
            return Image.open(requests.get(url, stream=True).raw)


def is_flash_attention_installed() -> (bool, str):
    try:
        import flash_attn
        return True, flash_attn.__version__
    except ImportError:
        return False, None


async def parse_request_body(
    request: Request,
) -> Tuple[Optional[Union[dict, str, bytes]], bool]:
    """
    Parse the request body while preserving it for future use.
    """
    try:
        # Ensure we have the raw body
        if not hasattr(request, '_body'):
            request._body = await request.body()
            async def get_body():
                return request._body
            request.body = get_body

        content_type = request.headers.get("Content-Type", "").lower()
        body = request._body

        if not body:
            return {}, False

        # Handle multipart form data
        if "multipart/form-data" in content_type:
            return body, True

        # Handle JSON content
        if "application/json" in content_type:
            try:
                return json.loads(body.decode("utf-8")), False
            except json.JSONDecodeError as e:
                logger.error(f"JSON decode error: {e}")
                return body, False

        # Handle text content
        elif "text/" in content_type:
            return body.decode("utf-8"), False

        # For other content types, return raw bytes
        return body, False

    except Exception as e:
        logger.error(f"Error parsing request body: {e}", exc_info=True)
        return {}, False

async def get_model_from_body(request: Request) -> str:
    """
    Extract model information from request body, handling different request types.

    Args:
        request (Request): The incoming HTTP request.

    Returns:
        str: The model name if found, empty string otherwise.
    """
    try:
        body, is_multipart = await parse_request_body(request)

        # For multipart requests (like audio), return default or configured model
        if is_multipart:
            return "whisper"  # Or any default model for audio

        # For regular JSON requests, extract model from body
        if isinstance(body, dict):
            return body.get("model", "")

        return ""

    except Exception as e:
        logger.error(f"Error while getting model from request: {e}")
        return ""


import pickle
import os
import re
import wordsegment
from g2p_en import G2p

from text.symbols import punctuation

from text.symbols2 import symbols

import unicodedata
from builtins import str as unicode
from g2p_en.expand import normalize_numbers
from nltk.tokenize import TweetTokenizer
word_tokenize = TweetTokenizer().tokenize
from nltk import pos_tag

current_file_path = os.path.dirname(__file__)
CMU_DICT_PATH = os.path.join(current_file_path, "cmudict.rep")
CMU_DICT_FAST_PATH = os.path.join(current_file_path, "cmudict-fast.rep")
CMU_DICT_HOT_PATH = os.path.join(current_file_path, "engdict-hot.rep")
CACHE_PATH = os.path.join(current_file_path, "engdict_cache.pickle")
NAMECACHE_PATH = os.path.join(current_file_path, "namedict_cache.pickle")

arpa = {
    "AH0",
    "S",
    "AH1",
    "EY2",
    "AE2",
    "EH0",
    "OW2",
    "UH0",
    "NG",
    "B",
    "G",
    "AY0",
    "M",
    "AA0",
    "F",
    "AO0",
    "ER2",
    "UH1",
    "IY1",
    "AH2",
    "DH",
    "IY0",
    "EY1",
    "IH0",
    "K",
    "N",
    "W",
    "IY2",
    "T",
    "AA1",
    "ER1",
    "EH2",
    "OY0",
    "UH2",
    "UW1",
    "Z",
    "AW2",
    "AW1",
    "V",
    "UW2",
    "AA2",
    "ER",
    "AW0",
    "UW0",
    "R",
    "OW1",
    "EH1",
    "ZH",
    "AE0",
    "IH2",
    "IH",
    "Y",
    "JH",
    "P",
    "AY1",
    "EY0",
    "OY2",
    "TH",
    "HH",
    "D",
    "ER0",
    "CH",
    "AO1",
    "AE1",
    "AO2",
    "OY1",
    "AY2",
    "IH1",
    "OW0",
    "L",
    "SH",
}


def replace_phs(phs):
    rep_map = {"'": "-"}
    phs_new = []
    for ph in phs:
        if ph in symbols:
            phs_new.append(ph)
        elif ph in rep_map.keys():
            phs_new.append(rep_map[ph])
        else:
            print("ph not in symbols: ", ph)
    return phs_new


def replace_consecutive_punctuation(text):
    punctuations = ''.join(re.escape(p) for p in punctuation)
    pattern = f'([{punctuations}])([{punctuations}])+'
    result = re.sub(pattern, r'\1', text)
    return result


def read_dict():
    g2p_dict = {}
    start_line = 49
    with open(CMU_DICT_PATH) as f:
        line = f.readline()
        line_index = 1
        while line:
            if line_index >= start_line:
                line = line.strip()
                word_split = line.split("  ")
                word = word_split[0].lower()

                syllable_split = word_split[1].split(" - ")
                g2p_dict[word] = []
                for syllable in syllable_split:
                    phone_split = syllable.split(" ")
                    g2p_dict[word].append(phone_split)

            line_index = line_index + 1
            line = f.readline()

    return g2p_dict


def read_dict_new():
    g2p_dict = {}
    with open(CMU_DICT_PATH) as f:
        line = f.readline()
        line_index = 1
        while line:
            if line_index >= 57:
                line = line.strip()
                word_split = line.split("  ")
                word = word_split[0].lower()
                g2p_dict[word] = [word_split[1].split(" ")]

            line_index = line_index + 1
            line = f.readline()

    with open(CMU_DICT_FAST_PATH) as f:
        line = f.readline()
        line_index = 1
        while line:
            if line_index >= 0:
                line = line.strip()
                word_split = line.split(" ")
                word = word_split[0].lower()
                if word not in g2p_dict:
                    g2p_dict[word] = [word_split[1:]]

            line_index = line_index + 1
            line = f.readline()

    return g2p_dict

def hot_reload_hot(g2p_dict):
    with open(CMU_DICT_HOT_PATH) as f:
        line = f.readline()
        line_index = 1
        while line:
            if line_index >= 0:
                line = line.strip()
                word_split = line.split(" ")
                word = word_split[0].lower()
                # 自定义发音词直接覆盖字典
                g2p_dict[word] = [word_split[1:]]

            line_index = line_index + 1
            line = f.readline()

    return g2p_dict


def cache_dict(g2p_dict, file_path):
    with open(file_path, "wb") as pickle_file:
        pickle.dump(g2p_dict, pickle_file)


def get_dict():
    if os.path.exists(CACHE_PATH):
        with open(CACHE_PATH, "rb") as pickle_file:
            g2p_dict = pickle.load(pickle_file)
    else:
        g2p_dict = read_dict_new()
        cache_dict(g2p_dict, CACHE_PATH)

    g2p_dict = hot_reload_hot(g2p_dict)

    return g2p_dict


def get_namedict():
    if os.path.exists(NAMECACHE_PATH):
        with open(NAMECACHE_PATH, "rb") as pickle_file:
            name_dict = pickle.load(pickle_file)
    else:
        name_dict = {}

    return name_dict


def text_normalize(text):
    # todo: eng text normalize
    # 适配中文及 g2p_en 标点
    rep_map = {
        "[;:：，；]": ",",
        '["’]': "'",
        "。": ".",
        "！": "!",
        "？": "?",
    }
    for p, r in rep_map.items():
        text = re.sub(p, r, text)

    # 来自 g2p_en 文本格式化处理
    # 增加大写兼容
    text = unicode(text)
    text = normalize_numbers(text)
    text = ''.join(char for char in unicodedata.normalize('NFD', text)
                    if unicodedata.category(char) != 'Mn')  # Strip accents
    text = re.sub("[^ A-Za-z'.,?!\-]", "", text)
    text = re.sub(r"(?i)i\.e\.", "that is", text)
    text = re.sub(r"(?i)e\.g\.", "for example", text)

    # 避免重复标点引起的参考泄露
    text = replace_consecutive_punctuation(text)

    return text


class en_G2p(G2p):
    def __init__(self):
        super().__init__()
        # 分词初始化
        wordsegment.load()

        # 扩展过时字典, 添加姓名字典
        self.cmu = get_dict()
        self.namedict = get_namedict()

        # 剔除读音错误的几个缩写
        for word in ["AE", "AI", "AR", "IOS", "HUD", "OS"]:
            del self.cmu[word.lower()]

        # 修正多音字
        self.homograph2features["read"] = (['R', 'IY1', 'D'], ['R', 'EH1', 'D'], 'VBP')
        self.homograph2features["complex"] = (['K', 'AH0', 'M', 'P', 'L', 'EH1', 'K', 'S'], ['K', 'AA1', 'M', 'P', 'L', 'EH0', 'K', 'S'], 'JJ')


    def __call__(self, text):
        # tokenization
        words = word_tokenize(text)
        tokens = pos_tag(words)  # tuples of (word, tag)

        # steps
        prons = []
        for o_word, pos in tokens:
            # 还原 g2p_en 小写操作逻辑
            word = o_word.lower()

            if re.search("[a-z]", word) is None:
                pron = [word]
            # 先把单字母推出去
            elif len(word) == 1:
                # 单读 A 发音修正, 这里需要原格式 o_word 判断大写
                if o_word == "A":
                    pron = ['EY1']
                else:
                    pron = self.cmu[word][0]
            # g2p_en 原版多音字处理
            elif word in self.homograph2features:  # Check homograph
                pron1, pron2, pos1 = self.homograph2features[word]
                if pos.startswith(pos1):
                    pron = pron1
                # pos1比pos长仅出现在read
                elif len(pos) < len(pos1) and pos == pos1[:len(pos)]:
                    pron = pron1
                else:
                    pron = pron2
            else:
                # 递归查找预测
                pron = self.qryword(o_word)

            prons.extend(pron)
            prons.extend([" "])

        return prons[:-1]


    def qryword(self, o_word):
        word = o_word.lower()

        # 查字典, 单字母除外
        if len(word) > 1 and word in self.cmu:  # lookup CMU dict
            return self.cmu[word][0]

        # 单词仅首字母大写时查找姓名字典
        if o_word.istitle() and word in self.namedict:
            return self.namedict[word][0]

        # oov 长度小于等于 3 直接读字母
        if len(word) <= 3:
            phones = []
            for w in word:
                # 单读 A 发音修正, 此处不存在大写的情况
                if w == "a":
                    phones.extend(['EY1'])
                elif not w.isalpha():
                    phones.extend([w])
                else:
                    phones.extend(self.cmu[w][0])
            return phones

        # 尝试分离所有格
        if re.match(r"^([a-z]+)('s)$", word):
            phones = self.qryword(word[:-2])[:]
            # P T K F TH HH 无声辅音结尾 's 发 ['S']
            if phones[-1] in ['P', 'T', 'K', 'F', 'TH', 'HH']:
                phones.extend(['S'])
            # S Z SH ZH CH JH 擦声结尾 's 发 ['IH1', 'Z'] 或 ['AH0', 'Z']
            elif phones[-1] in ['S', 'Z', 'SH', 'ZH', 'CH', 'JH']:
                phones.extend(['AH0', 'Z'])
            # B D G DH V M N NG L R W Y 有声辅音结尾 's 发 ['Z']
            # AH0 AH1 AH2 EY0 EY1 EY2 AE0 AE1 AE2 EH0 EH1 EH2 OW0 OW1 OW2 UH0 UH1 UH2 IY0 IY1 IY2 AA0 AA1 AA2 AO0 AO1 AO2
            # ER ER0 ER1 ER2 UW0 UW1 UW2 AY0 AY1 AY2 AW0 AW1 AW2 OY0 OY1 OY2 IH IH0 IH1 IH2 元音结尾 's 发 ['Z']
            else:
                phones.extend(['Z'])
            return phones

        # 尝试进行分词，应对复合词
        comps = wordsegment.segment(word.lower())

        # 无法分词的送回去预测
        if len(comps)==1:
            return self.predict(word)

        # 可以分词的递归处理
        return [phone for comp in comps for phone in self.qryword(comp)]


_g2p = en_G2p()


def g2p(text):
    # g2p_en 整段推理，剔除不存在的arpa返回
    phone_list = _g2p(text)
    phones = [ph if ph != "<unk>" else "UNK" for ph in phone_list if ph not in [" ", "<pad>", "UW", "</s>", "<s>"]]

    return replace_phs(phones)


if __name__ == "__main__":
    print(g2p("hello"))
    print(g2p(text_normalize("e.g. I used openai's AI tool to draw a picture.")))
    print(g2p(text_normalize("In this; paper, we propose 1 DSPGAN, a GAN-based universal vocoder.")))



# reference: https://github.com/ORI-Muchim/MB-iSTFT-VITS-Korean/blob/main/text/korean.py

import re
from jamo import h2j, j2hcj
import ko_pron
from g2pk2 import G2p

from text.symbols2 import symbols 

# This is a list of Korean classifiers preceded by pure Korean numerals.
_korean_classifiers = '군데 권 개 그루 닢 대 두 마리 모 모금 뭇 발 발짝 방 번 벌 보루 살 수 술 시 쌈 움큼 정 짝 채 척 첩 축 켤레 톨 통'

# List of (hangul, hangul divided) pairs:
_hangul_divided = [(re.compile('%s' % x[0]), x[1]) for x in [
    # ('ㄳ', 'ㄱㅅ'),   # g2pk2, A Syllable-ending Rule
    # ('ㄵ', 'ㄴㅈ'),
    # ('ㄶ', 'ㄴㅎ'),
    # ('ㄺ', 'ㄹㄱ'),
    # ('ㄻ', 'ㄹㅁ'),
    # ('ㄼ', 'ㄹㅂ'),
    # ('ㄽ', 'ㄹㅅ'),
    # ('ㄾ', 'ㄹㅌ'),
    # ('ㄿ', 'ㄹㅍ'),
    # ('ㅀ', 'ㄹㅎ'),
    # ('ㅄ', 'ㅂㅅ'),
    ('ㅘ', 'ㅗㅏ'),
    ('ㅙ', 'ㅗㅐ'),
    ('ㅚ', 'ㅗㅣ'),
    ('ㅝ', 'ㅜㅓ'),
    ('ㅞ', 'ㅜㅔ'),
    ('ㅟ', 'ㅜㅣ'),
    ('ㅢ', 'ㅡㅣ'),
    ('ㅑ', 'ㅣㅏ'),
    ('ㅒ', 'ㅣㅐ'),
    ('ㅕ', 'ㅣㅓ'),
    ('ㅖ', 'ㅣㅔ'),
    ('ㅛ', 'ㅣㅗ'),
    ('ㅠ', 'ㅣㅜ')
]]

# List of (Latin alphabet, hangul) pairs:
_latin_to_hangul = [(re.compile('%s' % x[0], re.IGNORECASE), x[1]) for x in [
    ('a', '에이'),
    ('b', '비'),
    ('c', '시'),
    ('d', '디'),
    ('e', '이'),
    ('f', '에프'),
    ('g', '지'),
    ('h', '에이치'),
    ('i', '아이'),
    ('j', '제이'),
    ('k', '케이'),
    ('l', '엘'),
    ('m', '엠'),
    ('n', '엔'),
    ('o', '오'),
    ('p', '피'),
    ('q', '큐'),
    ('r', '아르'),
    ('s', '에스'),
    ('t', '티'),
    ('u', '유'),
    ('v', '브이'),
    ('w', '더블유'),
    ('x', '엑스'),
    ('y', '와이'),
    ('z', '제트')
]]

# List of (ipa, lazy ipa) pairs:
_ipa_to_lazy_ipa = [(re.compile('%s' % x[0], re.IGNORECASE), x[1]) for x in [
    ('t͡ɕ','ʧ'),
    ('d͡ʑ','ʥ'),
    ('ɲ','n^'),
    ('ɕ','ʃ'),
    ('ʷ','w'),
    ('ɭ','l`'),
    ('ʎ','ɾ'),
    ('ɣ','ŋ'),
    ('ɰ','ɯ'),
    ('ʝ','j'),
    ('ʌ','ə'),
    ('ɡ','g'),
    ('\u031a','#'),
    ('\u0348','='),
    ('\u031e',''),
    ('\u0320',''),
    ('\u0339','')
]]


def fix_g2pk2_error(text):
    new_text = ""
    i = 0
    while i < len(text) - 4:
        if (text[i:i+3] == 'ㅇㅡㄹ' or text[i:i+3] == 'ㄹㅡㄹ') and text[i+3] == ' ' and text[i+4] == 'ㄹ':
            new_text += text[i:i+3] + ' ' + 'ㄴ'
            i += 5
        else:
            new_text += text[i]
            i += 1

    new_text += text[i:]
    return new_text


def latin_to_hangul(text):
    for regex, replacement in _latin_to_hangul:
        text = re.sub(regex, replacement, text)
    return text


def divide_hangul(text):
    text = j2hcj(h2j(text))
    for regex, replacement in _hangul_divided:
        text = re.sub(regex, replacement, text)
    return text


def hangul_number(num, sino=True):
    '''Reference https://github.com/Kyubyong/g2pK'''
    num = re.sub(',', '', num)

    if num == '0':
        return '영'
    if not sino and num == '20':
        return '스무'

    digits = '123456789'
    names = '일이삼사오육칠팔구'
    digit2name = {d: n for d, n in zip(digits, names)}

    modifiers = '한 두 세 네 다섯 여섯 일곱 여덟 아홉'
    decimals = '열 스물 서른 마흔 쉰 예순 일흔 여든 아흔'
    digit2mod = {d: mod for d, mod in zip(digits, modifiers.split())}
    digit2dec = {d: dec for d, dec in zip(digits, decimals.split())}

    spelledout = []
    for i, digit in enumerate(num):
        i = len(num) - i - 1
        if sino:
            if i == 0:
                name = digit2name.get(digit, '')
            elif i == 1:
                name = digit2name.get(digit, '') + '십'
                name = name.replace('일십', '십')
        else:
            if i == 0:
                name = digit2mod.get(digit, '')
            elif i == 1:
                name = digit2dec.get(digit, '')
        if digit == '0':
            if i % 4 == 0:
                last_three = spelledout[-min(3, len(spelledout)):]
                if ''.join(last_three) == '':
                    spelledout.append('')
                    continue
            else:
                spelledout.append('')
                continue
        if i == 2:
            name = digit2name.get(digit, '') + '백'
            name = name.replace('일백', '백')
        elif i == 3:
            name = digit2name.get(digit, '') + '천'
            name = name.replace('일천', '천')
        elif i == 4:
            name = digit2name.get(digit, '') + '만'
            name = name.replace('일만', '만')
        elif i == 5:
            name = digit2name.get(digit, '') + '십'
            name = name.replace('일십', '십')
        elif i == 6:
            name = digit2name.get(digit, '') + '백'
            name = name.replace('일백', '백')
        elif i == 7:
            name = digit2name.get(digit, '') + '천'
            name = name.replace('일천', '천')
        elif i == 8:
            name = digit2name.get(digit, '') + '억'
        elif i == 9:
            name = digit2name.get(digit, '') + '십'
        elif i == 10:
            name = digit2name.get(digit, '') + '백'
        elif i == 11:
            name = digit2name.get(digit, '') + '천'
        elif i == 12:
            name = digit2name.get(digit, '') + '조'
        elif i == 13:
            name = digit2name.get(digit, '') + '십'
        elif i == 14:
            name = digit2name.get(digit, '') + '백'
        elif i == 15:
            name = digit2name.get(digit, '') + '천'
        spelledout.append(name)
    return ''.join(elem for elem in spelledout)


def number_to_hangul(text):
    '''Reference https://github.com/Kyubyong/g2pK'''
    tokens = set(re.findall(r'(\d[\d,]*)([\uac00-\ud71f]+)', text))
    for token in tokens:
        num, classifier = token
        if classifier[:2] in _korean_classifiers or classifier[0] in _korean_classifiers:
            spelledout = hangul_number(num, sino=False)
        else:
            spelledout = hangul_number(num, sino=True)
        text = text.replace(f'{num}{classifier}', f'{spelledout}{classifier}')
    # digit by digit for remaining digits
    digits = '0123456789'
    names = '영일이삼사오육칠팔구'
    for d, n in zip(digits, names):
        text = text.replace(d, n)
    return text


def korean_to_lazy_ipa(text):
    text = latin_to_hangul(text)
    text = number_to_hangul(text)
    text=re.sub('[\uac00-\ud7af]+',lambda x:ko_pron.romanise(x.group(0),'ipa').split('] ~ [')[0],text)
    for regex, replacement in _ipa_to_lazy_ipa:
        text = re.sub(regex, replacement, text)
    return text

_g2p=G2p()
def korean_to_ipa(text):
    text = latin_to_hangul(text)
    text = number_to_hangul(text)
    text = _g2p(text)
    text = fix_g2pk2_error(text)
    text = korean_to_lazy_ipa(text)
    return text.replace('ʧ','tʃ').replace('ʥ','dʑ')

def post_replace_ph(ph):
    rep_map = {
        "：": ",",
        "；": ",",
        "，": ",",
        "。": ".",
        "！": "!",
        "？": "?",
        "\n": ".",
        "·": ",",
        "、": ",",
        "...": "…",
        " ": "空",
    }
    if ph in rep_map.keys():
        ph = rep_map[ph]
    if ph in symbols:
        return ph
    if ph not in symbols:
        ph = "停"
    return ph

def g2p(text):
    text = latin_to_hangul(text)
    text = _g2p(text)
    text = divide_hangul(text)
    text = fix_g2pk2_error(text)
    text = re.sub(r'([\u3131-\u3163])$', r'\1.', text)
    # text = "".join([post_replace_ph(i) for i in text])
    text = [post_replace_ph(i) for i in text]
    return text



import torch


def get_model():
    import whisper

    model = whisper.load_model("small", device="cpu")

    return model.encoder


def get_content(model=None, wav_16k_tensor=None):
    from whisper import log_mel_spectrogram, pad_or_trim

    dev = next(model.parameters()).device
    mel = log_mel_spectrogram(wav_16k_tensor).to(dev)[:, :3000]
    # if torch.cuda.is_available():
    #     mel = mel.to(torch.float16)
    feature_len = mel.shape[-1] // 2
    assert mel.shape[-1] < 3000, "输入音频过长，只允许输入30以内音频"
    with torch.no_grad():
        feature = model(pad_or_trim(mel, 3000).unsqueeze(0))[
            :1, :feature_len, :
        ].transpose(1, 2)
    return feature



ARTIFACT_SYSTEM_PROMPT = """
Be a friendly and upbeat conversational assistant.
You are LLM in an STT to LLM to TTS system.

Remember these key points:
- Keep responses brief, clear, and natural-sounding (like casual speech)
- Prefer simple words over complex terminology
- Break longer responses into shorter sentences
- Pause naturally between main points

Speech Recognition Handling:
- User input comes from speech transcription and may contain errors
- Common issues include pronunciation-based mistakes from non-native English speakers
- If a word seems incorrect or unclear, politely ask for clarification like "Did you mean [likely word]?"
- When in doubt about the entire question, confirm by repeating it back

Interactive Elements:
- User may interrupt during your response with <user_interrupt>
- When interrupted, stop immediately and address the interruption
- Resume previous topic only if explicitly requested

Response Style:
- Your response will be converted to voice to playback to user by TTS.
- The TTS system can playback non English words. Hence no English phonetic spellings needed.
- For example: Instead of 猫比狗聪明 (máo bǐ gǒu cōng míng), simple say this: 猫比狗聪明
- Use conversational markers like "Well," "You know," "Right"
- Express agreement/understanding with brief acknowledgments
- Mirror the user's speaking pace and energy level
- If technical terms are needed, define them in simple language

Error Recovery:
- If user seems confused, offer to rephrase or explain differently
- Acknowledge when you might have misunderstood something
- Be proactive in suggesting corrections for obvious transcription errors
"""



# ARTIFACT_SYSTEM_PROMPT = """
# Artifact System instruction:
# The assistant can create and reference artifacts during conversations. Artifacts are for substantial, self-contained content that users might modify or reuse, displayed in a separate UI window for clarity.
#
# Guidelines for structuring your response:
#
# # Good artifacts are:
# - Substantial content (>10 lines)
# - Content that the user is likely to modify, iterate on, or take ownership of
# - Self-contained, complex content that can be understood on its own, without context from the conversation
# - Content intended for eventual use outside the conversation (e.g., reports, emails, presentations)
# - Content likely to be referenced or reused multiple times
#
# # Don't use artifacts for:
# - Simple, informational, or short content, such as brief code snippets, mathematical equations, or small examples
# - Primarily explanatory, instructional, or illustrative content, such as examples provided to clarify a concept
# - Suggestions, commentary, or feedback on existing artifacts
# - Conversational or explanatory content that doesn't represent a standalone piece of work
# - Content that is dependent on the current conversational context to be useful
# - Content that is unlikely to be modified or iterated upon by the user
# - Request from users that appears to be a one-off question
#
# # Usage notes
# - One artifact per message unless specifically requested
# - Prefer in-line content (don't use artifacts) when possible. Unnecessary use of artifacts can be jarring for users.
# - If a user asks the assistant to "draw an SVG" or "make a website," the assistant does not need to explain that it doesn't have these capabilities. Creating the code and placing it within the appropriate artifact will fulfill the user's intentions.
# - If asked to generate an image, the assistant can offer an SVG instead. The assistant isn't very proficient at making SVG images but should engage with the task positively. Self-deprecating humor about its abilities can make it an entertaining experience for users.
# - The assistant errs on the side of simplicity and avoids overusing artifacts for content that can be effectively presented within the conversation.
#
# artifact_instructions:
#
# Structure your response according to these guidelines:
#
# 1. When using artifacts (default for substantial content):
# <answer>
#     <text>[Simple text here]</text>
#     <artifact identifier="[unique-identifier]" type="[artifact_type]" language="[language if applicable]" title="[Brief title]">
#     [Your content here]
#     </artifact>
#     <!-- Additional text or artifact as needed -->
# </answer>
#
# 2.When explicitly asked NOT to use artifacts or for simple responses:
# <answer>
#     <text>[Answer here]</text>
# </answer>
#
# 3. Use the <text> or <artifact> element for each distinct part of your response into chunks where applicable.
# 4. Use the <text> for content that doesnt form an artifact by the good artifacts criteria mentioned above..
# 5. For <artifact> elements:
#    - The "type" attribute should be one of: "code", or "self_contained_text".
#    - Use "code" for code snippets or examples.
#    - Use "self_contained_text" for detailed explanations or longer text content, standalone/self-contained elements.
#    - Include a unique "identifier" attribute using kebab-case (e.g., "example-code-snippet").
#    - Include a "language" attribute for code snippets (e.g., language="python").
#    - Include a brief "title" attribute describing the content.
# 6. Place your content directly within the <artifact> tags.
# 7. For code snippets, preserve indentation and line breaks.
# 8. You do not need to escape the content inside the <artifact> tags, as there will be a way to parse and handle that.
# 9. Both <text> and <artifact> elements of type=self_contained_text will be displayed in a separate UI window with markdown support.
# 10. If write code inside <text> tag please use ``` to indicate the codeblock and programming language.
#     For code inside <artifact>, do not use ``` to indicate.
#
# Example response structure when using artifact (default):
# <answer>
#     <text>Here is a brief explanation...</text>
#     <artifact identifier="example-function" type="code" language="python" title="Example Python Function">
# def example_function():
#     # code here
#     print("Hello, world!")
#     </artifact>
#     <artifact identifier="detailed-report" type="self_contained_text" title="Comprehensive Topic Analysis">
# A detailed report on the topic, which may include multiple paragraphs...
#     </artifact>
# </answer>
#
# Example response structure when user explicitly ask to not artifact:
# <answer>
#     <text>Here is a brief explanation...
#     ```python
#     # code here
#     print("Hello, world!")
#     ```
#     A detailed report on the topic, which may include multiple paragraphs...
#     </text>
# </answer>
#
# Remember: CDATA is prohibited. The data is processed in a special way and NO XML escaping is needed.
# XML comment ( <!-- tag) is prohibited.
# <text> and <self_contain_text> will be displayed in a separate UI window using markdown.
# Write content directly within the tags.
# The XML interpreter only recognize tag above (<answer>, <text>, <artifact>).
#
# Always precede an artifact by acknowledging the user's question or intention using <text>, unless it is unnecessary or explicitly requested other wise.
# End of Artifact System instruction.
#
# """


from .server_dataclass import (
    ModelRequest,
    ModelInstanceInfo,
    ModelInfo,
    AgentWithThinking,
    MixtureOfAgents,
    StopModelByPort
)

from .data_class import (
    BaseMessage,
    ChatMLQuery,
    ToolSpec,
    ToolForce,
    FunctionCall,
    ChoiceDeltaToolCall,
    GenerateQuery,
    ModelObjectResponse,
    ModelObject,
    ModelSpec,
    EmbeddingRequest,
    ModelDownloadSpec,
    SUPPORTED_BACKENDS,
    ChatCompletionResponse,
    UsageResponse,
    MultiModalTextContent,
    MultiModalImageContent
)

from .audio_data_class import (
    TranscriptionResponse,
    TimeStampedWord,
    TTSRequest,
    LanguageType
)
from .internal_ws import (
    WSInterTTS,
    TTSEvent,
    WSInterConfigUpdate,
    WSInterSTTResponse,
    WSInterCancel,
    WSInterCleanup
)

from .generation_data_class import (
    GenerationStats,
    GenStart,
    GenEnd,
    GenQueue,
    GenText,
    QueueContext,
    GenQueueDynamic
)


from .video import VideoFrameCollection, VideoFrame